// use std::cmp::Reverse;
use std::cell::RefCell;
use std::fmt::Debug;
use std::ops::Add;
use std::rc::Rc;

use crate::data_structures::graphs::edge::{EdgeTrait, HopLengthTrait};
use crate::data_structures::graphs::graph::{
    DoubleSidedGraphTrait, Graph, GraphTrait, SingleSideGraph,
};
use crate::data_structures::graphs::weight::{UnitWeight, WeightTrait};
use crate::data_structures::heaps::id_queue::MinIDQueue;
use crate::data_structures::timestampflags::TimeStampFlags;

use crate::algorithms::contraction_hierarchy::shortest_path_test::ShortestPathTest;

// use hashbrown::hash_map::DefaultHashBuilder;
// use priority_queue::PriorityQueue;
use std::time::Instant;

pub struct CHGraph<E, W>
where
    E: EdgeTrait<Weight = W> + HopLengthTrait + Debug,
    W: WeightTrait + Add<Output = W> + Debug,
{
    forward_graph: SingleSideGraph<E>,
    backward_graph: SingleSideGraph<E>,
    rank: Vec<u32>,
    order: Vec<u32>,
    level: Vec<u32>,
    max_pop_count: u32,
}

impl<E, W> CHGraph<E, W>
where
    E: EdgeTrait<Weight = W> + HopLengthTrait + Debug,
    W: WeightTrait + Add<Output = W> + Debug,
{
    pub fn num_vertices(&self) -> u32 {
        self.forward_graph.num_vertices()
    }

    fn estimate_vertex_importance(
        graph: Rc<RefCell<Graph<E>>>,
        shortest_path_test: Rc<RefCell<ShortestPathTest<E, W>>>,
        vertex: u32,
        level: u32,
    ) -> u32 {
        let mut added_arc_count = 0;
        let mut added_hop_count = 0;
        let mut total_vertices_settled = 0;

        // let graph = &(*((*shortest_path_test).graph));

        for i in 0..graph.borrow().in_degree(vertex) {
            let in_vertex = graph.borrow().inbound_edge(vertex, i).unwrap().to();
            let mut shortest_path_test_refmut = shortest_path_test.borrow_mut();
            shortest_path_test_refmut.pin_source(in_vertex, vertex);
            // println!("In = {}", in_vertex);

            for o in 0..graph.borrow().out_degree(vertex) {
                let out_vertex = graph.borrow().outbound_edge(vertex, o).unwrap().to();
                // println!("Out = {}", out_vertex);
                if in_vertex != out_vertex {
                    let (result, num_vertices_settled) = shortest_path_test_refmut
                        .does_shorter_or_equal_path_exist(
                            out_vertex,
                            graph
                                .borrow()
                                .inbound_edge(vertex, i)
                                .unwrap()
                                .weight()
                                .clone()
                                + graph
                                    .borrow()
                                    .outbound_edge(vertex, o)
                                    .unwrap()
                                    .weight()
                                    .clone(),
                        );

                    if !result {
                        added_arc_count += 1;
                        added_hop_count +=
                            graph.borrow().inbound_edge(vertex, i).unwrap().hop_length();
                        added_hop_count += graph
                            .borrow()
                            .outbound_edge(vertex, o)
                            .unwrap()
                            .hop_length();
                    }

                    total_vertices_settled += num_vertices_settled;
                }
            }
        }
        // println!(
        // 	"Settled {} vertices, in degree of node = {}, out degree = {}",
        // 	total_vertices_settled,
        // 	(*((*shortest_path_test).graph)).in_degree(vertex),
        // 	(*((*shortest_path_test).graph)).out_degree(vertex)
        // );

        let mut removed_arc_count: u32 = 1;
        removed_arc_count += graph.borrow().in_degree(vertex);
        removed_arc_count += graph.borrow().out_degree(vertex);

        let mut removed_hop_count: u32 = 1;
        for in_arc in graph.borrow().inbound_edges(vertex) {
            removed_hop_count += in_arc.hop_length();
        }
        for out_arc in graph.borrow().outbound_edges(vertex) {
            removed_hop_count += out_arc.hop_length();
        }

        1 + 1000 * level
            + (1000 * added_arc_count) / removed_arc_count
            + (1000 * added_hop_count) / removed_hop_count
    }

    pub fn build(mut graph: Graph<E>, max_pop_count: u32) -> CHGraph<E, W> {
        let num_vertices = graph.num_vertices();
        let mut forward_graph = SingleSideGraph::<E>::new(num_vertices);
        let mut backward_graph = SingleSideGraph::<E>::new(num_vertices);
        let mut _level = vec![0; num_vertices as usize];
        let mut _rank = vec![0; num_vertices as usize];
        let mut _order = vec![0; num_vertices as usize];

        let graph_rc = Rc::new(RefCell::new(graph));

        let mut shortest_path_test = ShortestPathTest::new(Rc::clone(&graph_rc), max_pop_count);
        let shortest_path_test_rc = Rc::new(RefCell::new(shortest_path_test));

        let start = Instant::now();
        // Add all importances of vertices to the heap
        let mut node_heap = MinIDQueue::<UnitWeight<u32>>::new(num_vertices);
        for v in 0..num_vertices {
            node_heap.push((
                v,
                UnitWeight::new(CHGraph::estimate_vertex_importance(
                    Rc::clone(&graph_rc),
                    Rc::clone(&shortest_path_test_rc),
                    v,
                    0, // Because nodes haven't been contracted yet
                )),
            ));
        }
        // println!("Finished building queue. Took {:?}", start.elapsed());

        let mut contracted_node_count = 0;
        while !node_heap.is_empty() {
            let (node_being_contracted, node_importance) = node_heap.pop();
            // println!(
            // 	"{} nodes contracted.. Contracting node {} of importance {:?}.",
            // 	contracted_node_count, node_being_contracted, node_importance
            // );

            _rank[node_being_contracted as usize] = contracted_node_count;
            _order[contracted_node_count as usize] = node_being_contracted;

            // Make a list of the neighbours
            // Perf note: I've tried marking and unmarking a boolean vector called is_neighbour
            // of the size num_vertices. It is slower on all benchmarks!
            // This version is simpler and faster.
            let mut neighbours = graph_rc.borrow().inbound_neighbours(node_being_contracted);
            neighbours.extend(
                graph_rc
                    .borrow()
                    .outbound_neighbours(node_being_contracted)
                    .iter(),
            );
            neighbours.sort(); // This might be slow
            neighbours.dedup();

            // Add edges to the graph
            for e in graph_rc
                .borrow()
                .outbound_edges(node_being_contracted)
                .iter()
            {
                forward_graph.add_edge(node_being_contracted, e.clone());
                // printf!("Added forward edge {} -> {}", node_being_contracted, e.to());
            }

            for e in graph_rc
                .borrow()
                .inbound_edges(node_being_contracted)
                .iter()
            {
                backward_graph.add_edge(node_being_contracted, e.clone());
                // printf!("Added backward edge {} -> {}", node_being_contracted, e.to());
            }

            // Contract the node:
            // This needs unsafe code, because0
            // ShortestPathTest::new() and graph.inbound/outbound_edges both take immutable borrows of the graph object,
            // while add_edge() needs a mutable borrow of the object, thus mixing two different types of borrows of the same
            // object, which is not allowed by the borrow checker.
            // See:: https://words.steveklabnik.com/you-can-t-turn-off-the-borrow-checker-in-rust
            {
                //println!("trying to borrow mutably");
                //let mut graph_ref_mut = graph_rc.borrow_mut();
                //println!("Borrowed! {}", graph_ref_mut.num_vertices());

                for in_edge in graph_rc.borrow().inbound_edges(node_being_contracted) {
                    //println!("Trying to borrow shortest_path_test mutably.... _in = {}", in_edge.to());
                    // let mut shortest_path_test_refmut = shortest_path_test_rc.borrow_mut();
                    //println!("Borrowed shortest_path_test mutably!");
                    shortest_path_test_rc
                        .borrow_mut()
                        .pin_source(in_edge.to(), node_being_contracted);

                    for out_edge in graph_rc.borrow().outbound_edges(node_being_contracted) {
                        //println!("Checking for shortest paths to {}", out_edge.to());
                        let (result, _) = shortest_path_test_rc
                            .borrow_mut()
                            .does_shorter_or_equal_path_exist(
                                out_edge.to().clone(),
                                in_edge.weight().clone() + out_edge.weight().clone(),
                            );

                        if !result {
                            if in_edge.to() != out_edge.to() {
                                // println!(
                                // 	"Adding edge {} -> {}, {:?}",
                                // 	in_edge.to(),
                                // 	out_edge.to(),
                                // 	in_edge.weight().clone() + out_edge.weight().clone()
                                // );

                                unsafe {
                                    let mut graph_ref_mut = graph_rc.as_ptr();
                                    (*graph_ref_mut).add_edge_or_reduce_weight(
                                        in_edge.to(),
                                        <E as HopLengthTrait>::new(
                                            out_edge.to(),
                                            in_edge.weight().clone() + out_edge.weight().clone(),
                                            in_edge.hop_length() + out_edge.hop_length(),
                                        ),
                                    );
                                }
                            }
                        }
                    }
                }
            }

            // 			if contracted_node_count % 10000 == 0 {
            // 				let out_degree = graph_rc.borrow().out_degree(node_being_contracted);
            // 				let in_degree = graph_rc.borrow().in_degree(node_being_contracted);
            //
            // 				println!("Contracted node {} of importance {:?}, {} of {}. The in degree of last node was {} and out degree was {}.", node_being_contracted, node_importance, contracted_node_count, graph_rc.borrow().num_vertices(), in_degree, out_degree);
            // 			}

            let neighbour_level = _level.get(node_being_contracted as usize).unwrap() + 1; // Safe because the size is checked
            graph_rc
                .borrow_mut()
                .remove_all_incident_edges(node_being_contracted);

            // Readjust the neighbour list
            for neighbour in neighbours.iter() {
                _level[neighbour.clone() as usize] = neighbour_level;

                let old_importance = node_heap.get_key(neighbour.clone());
                let new_importance = UnitWeight::new(CHGraph::estimate_vertex_importance(
                    Rc::clone(&graph_rc),
                    Rc::clone(&shortest_path_test_rc),
                    neighbour.clone(),
                    neighbour_level,
                ));

                if new_importance < old_importance {
                    node_heap.decrease_key((neighbour.clone(), new_importance));
                } else if new_importance > old_importance {
                    node_heap.increase_key((neighbour.clone(), new_importance));
                }
            }

            contracted_node_count += 1;
            neighbours.clear();
        }

        // 		println!(
        // 			"Size of forward graph is {} vertices, {} edges.",
        // 			forward_graph.num_vertices(),
        // 			forward_graph.num_edges()
        // 		);
        // 		println!(
        // 			"Size of backward graph is {} vertices, {} edges.",
        // 			backward_graph.num_vertices(),
        // 			backward_graph.num_edges()
        // 		);

        CHGraph {
            forward_graph,
            backward_graph,
            order: _order,
            rank: _rank,
            level: _level,
            max_pop_count,
        }
    }

    pub fn print(&self) {
        println!("=================== FORWARD GRAPH ====================");
        self.forward_graph.print();

        println!("=================== BACKWARD GRAPH ====================");
        self.backward_graph.print();

        //		println!("Order: ");
        //		for i in self.order.iter() {
        //			print!("{} ", i);
        //		}
        //
        //		println!("Rank: ");
        //		for i in self.rank.iter() {
        //			print!("{} ", i);
        //		}
    }

    pub fn print_edges(&self) {
        println!("=================== FORWARD GRAPH ====================");
        self.forward_graph.print_edges();

        println!("=================== BACKWARD GRAPH ====================");
        self.backward_graph.print_edges();
    }
}

pub struct CHQuery<'g, E, W>
where
    E: EdgeTrait<Weight = W> + HopLengthTrait + Debug,
    W: WeightTrait + Add<Output = W> + Debug,
{
    chgraph: &'g CHGraph<E, W>,
    forward_heap: MinIDQueue<W>,
    backward_heap: MinIDQueue<W>,
    forward_tentative_distance: Vec<W>,
    backward_tentative_distance: Vec<W>,
    was_forward_pushed: TimeStampFlags,
    was_backward_pushed: TimeStampFlags,
    parent_forward: Vec<u32>,
    parent_backward: Vec<u32>,
    shortest_path_meeting_node: Option<u32>,
}

impl<'g, E, W> CHQuery<'g, E, W>
where
    E: EdgeTrait<Weight = W> + HopLengthTrait + Debug,
    W: WeightTrait + Add<Output = W> + Debug,
{
    pub fn new(chgraph: &'g CHGraph<E, W>) -> CHQuery<'g, E, W> {
        let num_vertices = chgraph.num_vertices();
        CHQuery {
            chgraph,
            forward_heap: MinIDQueue::<W>::new(num_vertices),
            backward_heap: MinIDQueue::<W>::new(num_vertices),
            forward_tentative_distance: vec![W::MAX(); num_vertices as usize],
            backward_tentative_distance: vec![W::MAX(); num_vertices as usize],
            was_forward_pushed: TimeStampFlags::new(num_vertices as usize),
            was_backward_pushed: TimeStampFlags::new(num_vertices as usize),
            parent_forward: vec![0; num_vertices as usize],
            parent_backward: vec![0; num_vertices as usize],
            shortest_path_meeting_node: None,
        }
    }

    pub fn reset(&mut self) {
        self.forward_heap.clear();
        self.backward_heap.clear();
        self.was_forward_pushed.reset_all();
        self.was_backward_pushed.reset_all();
        self.shortest_path_meeting_node = None;
    }

    pub fn run(&mut self, from: u32, to: u32) -> Option<(W, u32)> {
        if from == to {
            return None;
        }

        self.forward_heap.push((from, W::ZERO()));
        self.forward_tentative_distance[from as usize] = W::ZERO();
        self.was_forward_pushed.set(from as usize);

        self.backward_heap.push((to, W::ZERO()));
        self.backward_tentative_distance[to as usize] = W::ZERO();
        self.was_backward_pushed.set(to as usize);

        let mut best_path_distance = W::MAX();
        let mut forward_settle = true;
        let mut settled_vertices_count = 0;

        #[allow(unreachable_code)]
        loop {
            let mut forward_finished = false;
            if self.forward_heap.is_empty() {
                forward_finished = true;
            } else if self.forward_heap.peek().unwrap().1 >= best_path_distance {
                forward_finished = true;
            }

            let mut backward_finished = false;
            if self.backward_heap.is_empty() {
                backward_finished = true;
            } else if self.backward_heap.peek().unwrap().1 >= best_path_distance {
                backward_finished = true;
            }

            if forward_finished && backward_finished {
                return match self.shortest_path_meeting_node {
                    None => None,
                    Some(meeting_node) => Some((
                        self.forward_tentative_distance[meeting_node as usize].clone()
                            + self.backward_tentative_distance[meeting_node as usize].clone(),
                        settled_vertices_count,
                    )),
                };
            }
            if forward_finished {
                forward_settle = false;
            }
            if backward_finished {
                forward_settle = true;
            }

            if forward_settle {
                forward_settle = false;
                settled_vertices_count += 1;

                let (popped_node, popped_node_distance) = self.forward_heap.pop();

                // If the popped node has been seen in the backward search,
                // Update the best path distance
                if self.was_backward_pushed.is_set(popped_node as usize) {
                    if best_path_distance
                        > popped_node_distance.clone()
                            + self.backward_tentative_distance[popped_node as usize].clone()
                    {
                        // println!("Best path distance updated forward!");
                        best_path_distance = popped_node_distance.clone()
                            + self.backward_tentative_distance[popped_node as usize].clone();
                        self.shortest_path_meeting_node = Some(popped_node);
                    }
                }

                // Stalling Technique:
                // Have I reached the popped_vertex from a path that is not the optimal path to popped_vertex?
                // If yes, i don't need to look further as all edges hereon are going to be at a distance
                // greater than the shortest distance possible from the source.
                let mut can_stall = false;
                for edge in self.chgraph.backward_graph.outbound_edges(popped_node) {
                    let next_node = edge.to();
                    if self.was_forward_pushed.is_set(next_node as usize) {
                        unsafe {
                            if (self
                                .forward_tentative_distance
                                .get_unchecked_mut(next_node as usize)
                                .clone()
                                + edge.weight().clone())
                                <= *self
                                    .forward_tentative_distance
                                    .get_unchecked_mut(popped_node as usize)
                            {
                                can_stall = true;
                                break;
                            }
                        }
                    }
                }

                // If I cannot stall, relax the edge to the popped vertex,
                // adding the neighbour vertices of popped vertex to the forward priority queue,
                // and adjusting the estimated distance to the neighbour vertices otherwise.
                if !can_stall {
                    for edge in self.chgraph.forward_graph.outbound_edges(popped_node) {
                        let next_node = edge.to();
                        let next_node_distance =
                            popped_node_distance.clone() + edge.weight().clone();

                        if self.was_forward_pushed.is_set(next_node as usize) {
                            if next_node_distance
                                < self.forward_tentative_distance[next_node as usize]
                            {
                                self.forward_heap
                                    .decrease_key((next_node, next_node_distance.clone()));
                                self.forward_tentative_distance[next_node as usize] =
                                    next_node_distance;
                            }
                        } else if next_node_distance < W::MAX() {
                            // println!("Pushing node {} to forward heap at a distance {:?}", next_node, next_node_distance);
                            self.forward_heap
                                .push((next_node, next_node_distance.clone()));
                            self.forward_tentative_distance[next_node as usize] =
                                next_node_distance;
                            self.was_forward_pushed.set(next_node as usize);
                        }
                    }
                }
            } else {
                forward_settle = true;
                settled_vertices_count += 1;

                let (popped_node, popped_node_distance) = self.backward_heap.pop();
                // println!("Popped backward node {} at distance {:?}", popped_node, popped_node_distance);

                if self.was_forward_pushed.is_set(popped_node as usize) {
                    if best_path_distance
                        > (popped_node_distance.clone()
                            + self.forward_tentative_distance[popped_node as usize].clone())
                    {
                        best_path_distance = popped_node_distance.clone()
                            + self.forward_tentative_distance[popped_node as usize].clone();
                        self.shortest_path_meeting_node = Some(popped_node);
                    }
                }

                let mut can_stall = false;
                for edge in self.chgraph.forward_graph.outbound_edges(popped_node) {
                    let next_node = edge.to();
                    if self.was_backward_pushed.is_set(next_node as usize) {
                        unsafe {
                            if (self
                                .backward_tentative_distance
                                .get_unchecked_mut(next_node as usize)
                                .clone()
                                + edge.weight().clone())
                                <= *self
                                    .backward_tentative_distance
                                    .get_unchecked_mut(popped_node as usize)
                            {
                                can_stall = true;
                                break;
                            }
                        }
                    }
                }

                if !can_stall {
                    for edge in self.chgraph.backward_graph.outbound_edges(popped_node) {
                        let next_node = edge.to();
                        let next_node_distance =
                            popped_node_distance.clone() + edge.weight().clone();

                        if self.was_backward_pushed.is_set(next_node as usize) {
                            if next_node_distance
                                < self.backward_tentative_distance[next_node as usize]
                            {
                                self.backward_heap
                                    .decrease_key((next_node, next_node_distance.clone()));
                                self.backward_tentative_distance[next_node as usize] =
                                    next_node_distance;
                            }
                        } else if next_node_distance < W::MAX() {
                            // println!("Pushing node {} to backward heap at a distance {:?}", next_node, next_node_distance);
                            self.backward_heap
                                .push((next_node, next_node_distance.clone()));
                            self.backward_tentative_distance[next_node as usize] =
                                next_node_distance;
                            self.was_backward_pushed.set(next_node as usize);
                        }
                    }
                }
            }
        }

        debug!(
            "Could not find a path, exiting, forward heap size = {}, back heap size = {}",
            self.forward_heap.size(),
            self.backward_heap.size()
        );
        None
    }
}
