// This file contains boolean versions of the dijkstra's algorithm, used for CH & EH
// Also, this is basically a Rust adaptation of
// RoutingKit's ShortestPathTest class found in contraction_hierarchy.cpp

use std::cell::RefCell;
use std::fmt::Debug;
use std::ops::Add;
use std::rc::Rc;

use crate::data_structures::graphs::edge::EdgeTrait;
use crate::data_structures::graphs::graph::{DoubleSidedGraphTrait, Graph, GraphTrait};
use crate::data_structures::graphs::weight::WeightTrait;
use crate::data_structures::heaps::id_queue::MinIDQueue;
use crate::data_structures::timestampflags::TimeStampFlags;

pub struct ShortestPathTest<E, W>
where
	E: EdgeTrait<Weight = W> + Debug,
	W: WeightTrait + Add<Output = W> + Debug,
{
	pub graph: Rc<RefCell<Graph<E>>>,
	forward_heap: MinIDQueue<W>,
	backward_heap: MinIDQueue<W>,
	forward_tentative_distance: Vec<W>,
	backward_tentative_distance: Vec<W>,
	was_forward_pushed: TimeStampFlags,
	was_backward_pushed: TimeStampFlags,
	max_pop_count: u32,
	bypass_node: Option<u32>,
}

impl<E, W> ShortestPathTest<E, W>
where
	E: EdgeTrait<Weight = W> + Debug,
	W: WeightTrait + Add<Output = W> + Debug,
{
	pub fn new(graph: Rc<RefCell<Graph<E>>>, max_pop_count: u32) -> ShortestPathTest<E, W> {
		let mut num_vertices = 0;
		num_vertices = graph.borrow().num_vertices();

		ShortestPathTest {
			graph,
			forward_heap: MinIDQueue::<W>::new(num_vertices),
			backward_heap: MinIDQueue::<W>::new(num_vertices),
			was_backward_pushed: TimeStampFlags::new(num_vertices as usize),
			was_forward_pushed: TimeStampFlags::new(num_vertices as usize),
			forward_tentative_distance: vec![W::ZERO(); num_vertices as usize],
			backward_tentative_distance: vec![W::ZERO(); num_vertices as usize],
			max_pop_count,
			bypass_node: None,
		}
	}

	pub fn pin_source(&mut self, source: u32, bypass_node: u32) {
		self.was_forward_pushed.reset_all();
		self.forward_heap.clear();
		self.forward_heap.push((source, W::ZERO()));
		self.forward_tentative_distance[source as usize] = W::ZERO();
		self.was_forward_pushed.set(source as usize);
		self.bypass_node = Some(bypass_node);
	}

	pub fn pin_source_without_bypass_node(&mut self, source: u32) {
		self.was_forward_pushed.reset_all();
		self.forward_heap.clear();
		self.forward_heap.push((source, W::ZERO()));
		self.forward_tentative_distance[source as usize] = W::ZERO();
		self.was_forward_pushed.set(source as usize);
		self.bypass_node = None;
	}

	// WARNING: THE BACKWARD_TENTATIVE_DISTANCE is not set to it's value in Routingkit in this function.
	// This could be a potential cause of errors
	pub fn does_shorter_or_equal_path_exist(&mut self, target: u32, curr_dist: W) -> (bool, u32) {
		self.was_backward_pushed.reset_all();
		self.backward_heap.clear();
		self.backward_heap.push((target, W::ZERO()));
		self.backward_tentative_distance[target as usize] = W::ZERO();
		self.was_backward_pushed.set(target as usize);

		let mut num_settled_vertices = 0;

		if self.was_forward_pushed.is_set(target as usize) {
			if self.forward_tentative_distance[target as usize] <= curr_dist {
				return (true, num_settled_vertices);
			}
		}

		let mut witness_found = false;

		while !self.forward_heap.is_empty() && !self.backward_heap.is_empty() {
			if num_settled_vertices > self.max_pop_count {
				return (false, num_settled_vertices);
			}

			if self.forward_heap.peek().unwrap().1.clone() + self.backward_heap.peek().unwrap().1.clone() > curr_dist {
				return (false, num_settled_vertices);
			}

			if self.forward_heap.peek().unwrap().1 <= self.backward_heap.peek().unwrap().1 {
				let (popped_node, popped_node_distance) = self.forward_heap.pop();
				debug!(
					"Settling forward vertex {} at distance {:?}",
					popped_node, popped_node_distance
				);

				for edge in self.graph.borrow().outbound_edges(popped_node) {
					let next_node = edge.to();
					if self.bypass_node != None && self.bypass_node.unwrap() == next_node {
						continue;
					}
					let next_node_distance = popped_node_distance.clone() + edge.weight().clone();

					// WARNING: Casting next_node from u32 to u16 may cause overflow errors.
					if self.was_forward_pushed.is_set(next_node as usize) {
						if next_node_distance < self.forward_tentative_distance[next_node as usize] {
							self.forward_heap.decrease_key((next_node, next_node_distance.clone()));
							*self.forward_tentative_distance.get_mut(next_node as usize).unwrap() =
								next_node_distance.clone();

							if self.was_backward_pushed.is_set(next_node as usize) {
								if next_node_distance
									+ self
										.backward_tentative_distance
										.get(next_node as usize)
										.unwrap()
										.clone() <= curr_dist
								{
									witness_found = true;
									return (witness_found, num_settled_vertices);
								}
							}
						}
					} else {
						self.was_forward_pushed.set(next_node as usize);
						*self.forward_tentative_distance.get_mut(next_node as usize).unwrap() =
							next_node_distance.clone();
						self.forward_heap.push((next_node, next_node_distance.clone()));

						if self.was_backward_pushed.is_set(next_node as usize) {
							if next_node_distance
								+ self
									.backward_tentative_distance
									.get_mut(next_node as usize)
									.unwrap()
									.clone() <= curr_dist
							{
								witness_found = true;
								return (witness_found, num_settled_vertices);
							}
						}
					}
				}
			} else {
				let (popped_node, popped_node_distance) = self.backward_heap.pop();
				debug!(
					"Settling backward vertex {} at distance {:?}",
					popped_node, popped_node_distance
				);

				for edge in self.graph.borrow().inbound_edges(popped_node) {
					let next_node = edge.to();
					if self.bypass_node != None && self.bypass_node.unwrap() == next_node {
						continue;
					}
					let next_node_distance = popped_node_distance.clone() + edge.weight().clone();

					if self.was_backward_pushed.is_set(next_node as usize) {
						if next_node_distance < self.backward_tentative_distance[next_node as usize] {
							self.backward_heap.decrease_key((next_node, next_node_distance.clone()));
							*self.backward_tentative_distance.get_mut(next_node as usize).unwrap() =
								next_node_distance.clone();

							if self.was_forward_pushed.is_set(next_node as usize) {
								if next_node_distance
									+ self.forward_tentative_distance.get(next_node as usize).unwrap().clone()
									<= curr_dist
								{
									witness_found = true;
									return (witness_found, num_settled_vertices);
								}
							}
						}
					} else {
						self.was_backward_pushed.set(next_node as usize);
						*self.backward_tentative_distance.get_mut(next_node as usize).unwrap() =
							next_node_distance.clone();
						self.backward_heap.push((next_node, next_node_distance.clone()));

						if self.was_forward_pushed.is_set(next_node as usize) {
							if next_node_distance
								+ self
									.forward_tentative_distance
									.get_mut(next_node as usize)
									.unwrap()
									.clone() <= curr_dist
							{
								witness_found = true;
								return (witness_found, num_settled_vertices);
							}
						}
					}
				}
			}

			num_settled_vertices += 1;
		}

		//	debug!(
		//		"Could not find a path, exiting, forward heap size = {}, back heap size = {}",
		//		forward_heap.len(),
		//		backward_heap.len()
		//	);
		(witness_found, num_settled_vertices)
	}
}

// Returns the distance between from and to vertices if a path is found, None otherwise
// pub fn simple_shorter_or_equal_path_to_target_exists<E, W>(
// 	_graph: &Graph<E>,
// 	_from: usize,
// 	_to: usize,
// ) -> bool
// where
// 	E: EdgeTrait<Weight = W> + Debug,
// 	W: WeightTrait + Add<Output = W> + Debug,
// {
// 	// Trivial case
// 	if _from == _to {
// 		return true;
// 	}
//
// 	let mut heap = LabelHeap::<W>::new();
// 	let mut settled_vertices = HashSet::<usize>::new();
//
// 	heap.push(Label::with_weight(_from, W::ZERO()));
//
// 	while let Some(Label { _id, _distance }) = heap.pop() {
// 		// If reached the destination
// 		if _id == _to {
// 			return true;
// 		}
//
// 		// Settle the vertex
// 		settled_vertices.insert(_id);
//
// 		// Add all the neighbours in the heap
// 		for edge in _graph.outbound_edges(_id) {
// 			// If not already settled, add to the priority queue
// 			if !settled_vertices.contains(&edge.to()) {
// 				let new_distance_to_v =
// 					_distance.clone() + edge.weight().clone();
// 				heap.push(Label::with_weight(edge.to(), new_distance_to_v));
// 			}
// 		}
// 	}
//
// 	false
// }
//
// pub fn shorter_or_equal_path_to_target_exists<E, W>(
// 	_graph: &Graph<E>,
// 	_from: usize,
// 	_to: usize,
// 	_current_weight: W
// ) -> bool
// where
// 	E: EdgeTrait<Weight = W> + Debug,
// 	W: WeightTrait + Add<Output = W> + Debug,
// {
// 	false
// }
