use rustc_hash::{FxHashMap, FxHashSet};
use std::fmt::Debug;
use std::ops::Add;
use std::rc::Rc;

use crate::data_structures::graphs::edge::*;
use crate::data_structures::graphs::graph::{DoubleSidedGraphTrait, GraphTrait};
use crate::data_structures::graphs::weight::*;
use crate::data_structures::heaps::id_queue::MinIDQueue;
use crate::data_structures::heaps::label_priority_queue::*;

use bit_set::BitSet;

pub struct Dijkstra<E, W>
where
    E: EdgeTrait<Weight = W> + Debug,
    W: WeightTrait + Add<Output = W> + Debug,
{
    graph: Rc<dyn GraphTrait<E>>,
    heap: LabelHeap<W>,
    settled_vertices: Vec<u32>,
    settled_vertex_weights: Vec<W>,
    parent: Vec<u32>,
    settled_vertices_indexes: FxHashSet<u32>,
    settled_vertex_count: u32,
}

impl<E, W> Dijkstra<E, W>
where
    E: EdgeTrait<Weight = W> + Debug,
    W: WeightTrait + Add<Output = W> + Debug,
{
    pub fn new(graph: Rc<dyn GraphTrait<E>>) -> Dijkstra<E, W> {
        // println!("Created DIJKSTRA STRUCT");
        Dijkstra {
            heap: LabelHeap::<W>::new(),
            settled_vertices: vec![0; graph.num_vertices() as usize],
            settled_vertex_weights: vec![W::ZERO(); graph.num_vertices() as usize],
            parent: vec![0; graph.num_vertices() as usize],
            settled_vertices_indexes: FxHashSet::default(),
            settled_vertex_count: 0,
            graph,
        }
    }

    pub fn reset(&mut self) {
        // print!("RESET");
        self.heap.clear();

        for i in self.settled_vertices_indexes.iter() {
            self.settled_vertices[*i as usize] = 0;
            self.settled_vertex_weights[*i as usize] = W::ZERO();
            self.parent[*i as usize] = 0;
        }

        self.settled_vertices_indexes.clear();
        self.settled_vertex_count = 0;
    }

    pub fn run(&mut self, from: u32, to: u32) -> Option<(W, u32)> {
        // print!("Running...... ");
        if from == to {
            return None;
        }

        self.heap.push(Label::with_weight(from, W::ZERO()));

        while let Some(Label { id, distance }) = self.heap.pop() {
            self.settled_vertex_count += 1;

            debug!("Popped label {}", id);

            // If reached the destination
            if id == to {
                debug!("Found a path with distance {:?}", distance);
                return Some((distance, self.settled_vertex_count));
            }

            // Settle the vertex
            self.settled_vertices[id as usize] = 1;
            self.settled_vertices_indexes.insert(id);

            // Add all the neighbours in the heap
            for edge in self.graph.outbound_edges(id) {
                // If not already settled, add to the priority queue
                if self.settled_vertices[edge.to() as usize] == 0 {
                    debug!("Checking edge: {}", edge.to());
                    debug!("destination vertex not settled");

                    let new_distance_to_v = distance.clone() + edge.weight().clone();
                    self.heap
                        .push(Label::with_weight(edge.to(), new_distance_to_v));

                    //  debug!("Queue: {:?}", heap);
                }
            }
        }

        None
    }
}

// Returns the distance between from and to vertices if a path is found, None otherwise
pub fn dijkstra<E, W>(graph: Rc<dyn GraphTrait<E>>, from: u32, to: u32) -> Option<(W, u32)>
where
    E: EdgeTrait<Weight = W> + Debug,
    W: WeightTrait + Add<Output = W> + Debug,
{
    if from == to {
        return None;
    }

    let mut heap = LabelHeap::<W>::new();
    let mut settled_vertices = FxHashSet::default();

    heap.push(Label::with_weight(from, W::ZERO()));

    while let Some(Label { id, distance }) = heap.pop() {
        debug!("Popped label {}", id);

        // If reached the destination
        if id == to {
            debug!("Found a path with distance {:?}", distance);
            return Some((distance, settled_vertices.len() as u32));
        }

        // Settle the vertex
        settled_vertices.insert(id);

        // Add all the neighbours in the heap
        for edge in graph.outbound_edges(id) {
            // If not already settled, add to the priority queue
            if !settled_vertices.contains(&edge.to()) {
                debug!("Checking edge: {}", edge.to());
                debug!("destination vertex not settled");

                let new_distance_to_v = distance.clone() + edge.weight().clone();
                heap.push(Label::with_weight(edge.to(), new_distance_to_v));

                //  debug!("Queue: {:?}", heap);
            }
        }
    }

    None
}

pub struct BidirectionalDijkstra<E, W>
where
    E: EdgeTrait<Weight = W> + Debug,
    W: WeightTrait + Add<Output = W> + Debug,
{
    graph: Rc<dyn DoubleSidedGraphTrait<E>>,
    forward_heap: LabelHeap<W>,
    backward_heap: LabelHeap<W>,
    settled_vertices_forward: Vec<u32>,
    settled_vertices_backward: Vec<u32>,
    settled_vertices_forward_weights: Vec<W>,
    settled_vertices_backward_weights: Vec<W>,
    parent_forward: Vec<u32>,
    parent_backward: Vec<u32>,
    settled_vertices_forward_indexes: BitSet,
    settled_vertices_backward_indexes: BitSet,
    settled_vertices_forward_count: u32,
    settled_vertices_backward_count: u32,
}

impl<E, W> BidirectionalDijkstra<E, W>
where
    E: EdgeTrait<Weight = W> + Debug,
    W: WeightTrait + Add<Output = W> + Debug,
{
    pub fn new(graph: Rc<dyn DoubleSidedGraphTrait<E>>) -> BidirectionalDijkstra<E, W> {
        BidirectionalDijkstra {
            forward_heap: LabelHeap::<W>::new(),
            backward_heap: LabelHeap::<W>::new(),
            settled_vertices_forward: vec![0; graph.num_vertices() as usize],
            settled_vertices_backward: vec![0; graph.num_vertices() as usize],
            settled_vertices_forward_weights: vec![W::ZERO(); graph.num_vertices() as usize],
            settled_vertices_backward_weights: vec![W::ZERO(); graph.num_vertices() as usize],
            parent_forward: vec![0; graph.num_vertices() as usize],
            parent_backward: vec![0; graph.num_vertices() as usize],
            settled_vertices_forward_indexes: BitSet::with_capacity(graph.num_vertices() as usize),
            settled_vertices_backward_indexes: BitSet::with_capacity(graph.num_vertices() as usize),
            settled_vertices_forward_count: 0,
            settled_vertices_backward_count: 0,
            graph,
        }
    }

    pub fn reset(&mut self) {
        self.forward_heap.clear();
        self.backward_heap.clear();

        // Avoiding runtime bounds checking needs the unsafe blocks
        for i in self.settled_vertices_forward_indexes.iter() {
            unsafe {
                *self.settled_vertices_forward.get_unchecked_mut(i) = 0;
                *self.settled_vertices_forward_weights.get_unchecked_mut(i) = W::ZERO();
                *self.parent_forward.get_unchecked_mut(i) = 0;
            }
        }

        for i in self.settled_vertices_backward_indexes.iter() {
            unsafe {
                *self.settled_vertices_backward.get_unchecked_mut(i) = 0;
                *self.settled_vertices_backward_weights.get_unchecked_mut(i) = W::ZERO();
                *self.parent_backward.get_unchecked_mut(i) = 0;
            }
        }

        self.settled_vertices_forward_indexes.clear();
        self.settled_vertices_backward_indexes.clear();
        self.settled_vertices_forward_count = 0;
        self.settled_vertices_backward_count = 0;
    }

    pub fn run(&mut self, _from: u32, _to: u32) -> Option<(W, u32)> {
        if _from == _to {
            return None;
        }

        self.forward_heap.push(Label::with_weight(_from, W::ZERO()));
        self.backward_heap.push(Label::with_weight(_to, W::ZERO()));

        let mut best_path_distance = W::MAX();
        let mut forward_settle = true;

        while !self.forward_heap.is_empty() && !self.backward_heap.is_empty() {
            let front = self.forward_heap.peek().unwrap();
            let back = self.backward_heap.peek().unwrap();

            if best_path_distance != W::MAX()
                && front.distance.clone() + back.distance.clone() >= best_path_distance.clone()
            {
                return Some((
                    best_path_distance,
                    self.settled_vertices_forward_count + self.settled_vertices_backward_count,
                ));
            }

            if forward_settle {
                forward_settle = false;
                self.settled_vertices_forward_count += 1;

                let Label { id, distance } = self.forward_heap.pop().unwrap();
                //				println!(
                //					"Extracted forward vertex {} at distance {:?}",
                //					_id, _distance
                //				);
                unsafe {
                    *self.settled_vertices_forward.get_unchecked_mut(id as usize) = 1;
                    *self
                        .settled_vertices_forward_weights
                        .get_unchecked_mut(id as usize) = distance.clone();
                }
                self.settled_vertices_forward_indexes.insert(id as usize);

                for edge in self.graph.outbound_edges(id) {
                    let mut update_best_path_distance = |_distance_to_v: &W| {
                        let new_distance_to_v =
                            distance.clone() + edge.weight().clone() + _distance_to_v.clone();

                        if new_distance_to_v < best_path_distance.clone() {
                            best_path_distance = new_distance_to_v;
                        }
                    };

                    if self
                        .settled_vertices_backward
                        .get(edge.to() as usize)
                        .unwrap()
                        .clone()
                        == 1 as u32
                    {
                        update_best_path_distance(
                            self.settled_vertices_backward_weights
                                .get(edge.to() as usize)
                                .unwrap(),
                        );
                    } else if self.backward_heap.contains(&edge.to()) {
                        update_best_path_distance(
                            &self
                                .backward_heap
                                .best_known_distance_to_id(&edge.to())
                                .unwrap(),
                        );
                    }

                    if self
                        .settled_vertices_forward
                        .get(edge.to() as usize)
                        .unwrap()
                        .clone()
                        == 0 as u32
                    {
                        let new_distance_to_v = distance.clone() + edge.weight().clone();
                        //						println!("Pushing forward {}, {:?}", edge.to(), new_distance_to_v);
                        self.forward_heap
                            .push(Label::with_weight(edge.to(), new_distance_to_v));
                    }
                }
            } else {
                forward_settle = true;
                self.settled_vertices_backward_count += 1;

                let Label { id, distance } = self.backward_heap.pop().unwrap();
                //				println!("Extracted backward vertex {} at distance {:?}", _id, _distance);
                unsafe {
                    *self
                        .settled_vertices_backward
                        .get_unchecked_mut(id as usize) = 1;
                    *self
                        .settled_vertices_backward_weights
                        .get_unchecked_mut(id as usize) = distance.clone();
                }
                self.settled_vertices_backward_indexes.insert(id as usize);

                for edge in self.graph.inbound_edges(id) {
                    let mut update_best_path_distance = |_distance_to_v: &W| {
                        let new_distance_to_v =
                            distance.clone() + edge.weight().clone() + _distance_to_v.clone();

                        if new_distance_to_v < best_path_distance.clone() {
                            best_path_distance = new_distance_to_v;
                        }
                    };

                    if self
                        .settled_vertices_forward
                        .get(edge.to() as usize)
                        .unwrap()
                        .clone()
                        == 1 as u32
                    {
                        update_best_path_distance(
                            self.settled_vertices_forward_weights
                                .get(edge.to() as usize)
                                .unwrap(),
                        );
                    } else if self.forward_heap.contains(&edge.to()) {
                        update_best_path_distance(
                            &self
                                .forward_heap
                                .best_known_distance_to_id(&edge.to())
                                .unwrap(),
                        );
                    }

                    if self
                        .settled_vertices_backward
                        .get(edge.to() as usize)
                        .unwrap()
                        .clone()
                        == 0 as u32
                    {
                        let new_distance_to_v = distance.clone() + edge.weight().clone();
                        // println!("Pushing backward {}, {:?}", edge.to(), new_distance_to_v);
                        self.backward_heap
                            .push(Label::with_weight(edge.to(), new_distance_to_v));
                    }
                }
            }
        }

        //	debug!(
        //		"Could not find a path, exiting, forward heap size = {}, back heap size = {}",
        //		forward_heap.len(),
        //		backward_heap.len()
        //	);
        None
    }
}

pub struct BidirectionalDijkstraMinQ<E, W>
where
    E: EdgeTrait<Weight = W> + Debug,
    W: WeightTrait + Add<Output = W> + Debug,
{
    graph: Rc<dyn DoubleSidedGraphTrait<E>>,
    forward_heap: MinIDQueue<W>,
    backward_heap: MinIDQueue<W>,
    settled_vertices_forward: Vec<u32>,
    settled_vertices_backward: Vec<u32>,
    settled_vertices_forward_weights: Vec<W>,
    settled_vertices_backward_weights: Vec<W>,
    parent_forward: Vec<u32>,
    parent_backward: Vec<u32>,
    settled_vertices_forward_indexes: BitSet,
    settled_vertices_backward_indexes: BitSet,
    settled_vertices_forward_count: u32,
    settled_vertices_backward_count: u32,
}

impl<E, W> BidirectionalDijkstraMinQ<E, W>
where
    E: EdgeTrait<Weight = W> + Debug,
    W: WeightTrait + Add<Output = W> + Debug,
{
    pub fn new(graph: Rc<dyn DoubleSidedGraphTrait<E>>) -> BidirectionalDijkstraMinQ<E, W> {
        BidirectionalDijkstraMinQ {
            forward_heap: MinIDQueue::<W>::new(graph.num_vertices()),
            backward_heap: MinIDQueue::<W>::new(graph.num_vertices()),
            settled_vertices_forward: vec![0; graph.num_vertices() as usize],
            settled_vertices_backward: vec![0; graph.num_vertices() as usize],
            settled_vertices_forward_weights: vec![W::ZERO(); graph.num_vertices() as usize],
            settled_vertices_backward_weights: vec![W::ZERO(); graph.num_vertices() as usize],
            parent_forward: vec![0; graph.num_vertices() as usize],
            parent_backward: vec![0; graph.num_vertices() as usize],
            settled_vertices_forward_indexes: BitSet::with_capacity(graph.num_vertices() as usize),
            settled_vertices_backward_indexes: BitSet::with_capacity(graph.num_vertices() as usize),
            settled_vertices_forward_count: 0,
            settled_vertices_backward_count: 0,
            graph,
        }
    }

    pub fn reset(&mut self) {
        self.forward_heap.clear();
        self.backward_heap.clear();

        // Avoiding runtime bounds checking needs the unsafe blocks
        for i in self.settled_vertices_forward_indexes.iter() {
            unsafe {
                *self.settled_vertices_forward.get_unchecked_mut(i) = 0;
                *self.settled_vertices_forward_weights.get_unchecked_mut(i) = W::ZERO();
                *self.parent_forward.get_unchecked_mut(i) = 0;
            }
        }

        for i in self.settled_vertices_backward_indexes.iter() {
            unsafe {
                *self.settled_vertices_backward.get_unchecked_mut(i) = 0;
                *self.settled_vertices_backward_weights.get_unchecked_mut(i) = W::ZERO();
                *self.parent_backward.get_unchecked_mut(i) = 0;
            }
        }

        self.settled_vertices_forward_indexes.clear();
        self.settled_vertices_backward_indexes.clear();
        self.settled_vertices_forward_count = 0;
        self.settled_vertices_backward_count = 0;
    }

    pub fn run(&mut self, _from: u32, _to: u32) -> Option<(W, u32)> {
        if _from == _to {
            return None;
        }

        self.forward_heap.push((_from, W::ZERO()));
        self.backward_heap.push((_to, W::ZERO()));

        let mut best_path_distance = W::MAX();
        let mut forward_settle = true;

        while !self.forward_heap.is_empty() && !self.backward_heap.is_empty() {
            let front = self.forward_heap.peek().unwrap();
            let back = self.backward_heap.peek().unwrap();

            if best_path_distance != W::MAX()
                && front.1.clone() + back.1.clone() >= best_path_distance.clone()
            {
                //				println!("Found path, returning {:?}", best_path_distance);
                return Some((
                    best_path_distance,
                    self.settled_vertices_forward_count + self.settled_vertices_backward_count,
                ));
            }

            if forward_settle {
                forward_settle = false;
                self.settled_vertices_forward_count += 1;

                let (_id, _distance) = self.forward_heap.pop();
                debug!(
                    "Settling forward vertex {} at distance {:?}",
                    _id, _distance
                );
                unsafe {
                    *self
                        .settled_vertices_forward
                        .get_unchecked_mut(_id as usize) = 1;
                    *self
                        .settled_vertices_forward_weights
                        .get_unchecked_mut(_id as usize) = _distance.clone();
                }
                self.settled_vertices_forward_indexes.insert(_id as usize);

                for edge in self.graph.outbound_edges(_id) {
                    let mut update_best_path_distance = |_distance_to_v: &W| {
                        let new_distance_to_v =
                            _distance.clone() + edge.weight().clone() + _distance_to_v.clone();

                        if new_distance_to_v < best_path_distance.clone() {
                            best_path_distance = new_distance_to_v;
                        }
                    };

                    if self
                        .settled_vertices_backward
                        .get(edge.to() as usize)
                        .unwrap()
                        .clone()
                        == 1 as u32
                    {
                        update_best_path_distance(
                            self.settled_vertices_backward_weights
                                .get(edge.to() as usize)
                                .unwrap(),
                        );
                    } else if self.backward_heap.contains_id(edge.to()) {
                        update_best_path_distance(&self.backward_heap.get_key(edge.to()));
                    }

                    if self
                        .settled_vertices_forward
                        .get(edge.to() as usize)
                        .unwrap()
                        .clone()
                        == 0 as u32
                    {
                        let new_distance_to_v = _distance.clone() + edge.weight().clone();

                        if self.forward_heap.contains_id(edge.to()) {
                            if self.forward_heap.get_key(edge.to()) > new_distance_to_v {
                                self.forward_heap
                                    .decrease_key((edge.to(), new_distance_to_v));
                            }
                        } else {
                            //println!("Pushing forward {}, {:?}", edge.to(), new_distance_to_v);
                            self.forward_heap.push((edge.to(), new_distance_to_v));
                        }
                    }
                }
            } else {
                forward_settle = true;
                self.settled_vertices_backward_count += 1;

                let (_id, _distance) = self.backward_heap.pop();
                debug!(
                    "Settling backward vertex {} at distance {:?}",
                    _id, _distance
                );
                unsafe {
                    *self
                        .settled_vertices_backward
                        .get_unchecked_mut(_id as usize) = 1;
                    *self
                        .settled_vertices_backward_weights
                        .get_unchecked_mut(_id as usize) = _distance.clone();
                }
                self.settled_vertices_backward_indexes.insert(_id as usize);

                for edge in self.graph.inbound_edges(_id) {
                    let mut update_best_path_distance = |_distance_to_v: &W| {
                        let new_distance_to_v =
                            _distance.clone() + edge.weight().clone() + _distance_to_v.clone();

                        if new_distance_to_v < best_path_distance.clone() {
                            best_path_distance = new_distance_to_v;
                        }
                    };

                    if self
                        .settled_vertices_forward
                        .get(edge.to() as usize)
                        .unwrap()
                        .clone()
                        == 1 as u32
                    {
                        update_best_path_distance(
                            self.settled_vertices_forward_weights
                                .get(edge.to() as usize)
                                .unwrap(),
                        );
                    } else if self.forward_heap.contains_id(edge.to()) {
                        update_best_path_distance(&self.forward_heap.get_key(edge.to()));
                    }

                    if self
                        .settled_vertices_backward
                        .get(edge.to() as usize)
                        .unwrap()
                        .clone()
                        == 0 as u32
                    {
                        let new_distance_to_v = _distance.clone() + edge.weight().clone();
                        if self.backward_heap.contains_id(edge.to()) {
                            if self.backward_heap.get_key(edge.to()) > new_distance_to_v {
                                self.backward_heap
                                    .decrease_key((edge.to(), new_distance_to_v));
                            }
                        } else {
                            debug!("Pushing backward {}, {:?}", edge.to(), new_distance_to_v);
                            self.backward_heap.push((edge.to(), new_distance_to_v));
                        }
                    }
                }
            }
        }

        //	debug!(
        //		"Could not find a path, exiting, forward heap size = {}, back heap size = {}",
        //		forward_heap.len(),
        //		backward_heap.len()
        //	);
        None
    }
}

pub fn bidirectional_dijkstra<E, W>(
    graph: Rc<dyn DoubleSidedGraphTrait<E>>,
    from: u32,
    to: u32,
) -> Option<(W, u32)>
where
    E: EdgeTrait<Weight = W> + Debug,
    W: WeightTrait + Add<Output = W> + Debug,
{
    //	debug!("Starting bidirectional dijkstra's");

    if from == to {
        return None;
    }

    let mut forward_heap = LabelHeap::<W>::new();
    let mut backward_heap = LabelHeap::<W>::new();
    let mut settled_vertices_forward = FxHashMap::<u32, W>::default();
    let mut settled_vertices_backward = FxHashMap::<u32, W>::default();

    forward_heap.push(Label::with_weight(from, W::ZERO()));
    backward_heap.push(Label::with_weight(to, W::ZERO()));

    let mut best_path_distance = W::MAX();
    let mut forward_settle = true;
    let mut settled_vertices_count = 0;

    while !forward_heap.is_empty() && !backward_heap.is_empty() {
        settled_vertices_count += 1;

        // Safe to unwrap() since we just checked if they were empty
        let front = forward_heap.peek().unwrap();
        let back = backward_heap.peek().unwrap();

        // Stopping condition:
        // If the distance along this path is greater than the best known path
        // distance, stop searching for any more paths and return.
        if best_path_distance != W::MAX()
            && front.distance.clone() + back.distance.clone() >= best_path_distance.clone()
        {
            // Isn't this redundant?
            //						println!("Found path, returning {:?}", best_path_distance);
            return Some((best_path_distance, settled_vertices_count));
        }

        if forward_settle {
            forward_settle = false;

            // Extract a label from the forward Priority Queue
            let Label { id, distance } = forward_heap.pop().unwrap();

            //			debug!("Forward Extracted {} at distance {:?}", _id, _distance);

            // Settle the vertex
            settled_vertices_forward.insert(id, distance.clone());

            // Iterate over all the edges
            for edge in graph.outbound_edges(id) {
                //				debug!(
                //					"iterating over edge to {}, weight {:?}",
                //					edge.to(),
                //					edge.weight()
                //				);

                // Has the vertex been seen in the backward search yet?
                let mut update_best_path_distance = |_distance_to_v: &W| {
                    let new_distance_to_v =
                        distance.clone() + edge.weight().clone() + _distance_to_v.clone();

                    if new_distance_to_v < best_path_distance.clone() {
                        best_path_distance = new_distance_to_v;
                        //						debug!(
                        //							"Best path distance updated to {:?}",
                        //							best_path_distance
                        //						);
                    }
                };

                if settled_vertices_backward.contains_key(&edge.to()) {
                    update_best_path_distance(settled_vertices_backward.get(&edge.to()).unwrap());
                } else if backward_heap.contains(&edge.to()) {
                    update_best_path_distance(
                        &backward_heap.best_known_distance_to_id(&edge.to()).unwrap(),
                    );
                }

                // If the vertex hasn't been settled yet, add it to the forward heap
                if !settled_vertices_forward.contains_key(&edge.to()) {
                    //					debug!("Checking edge: {}", edge.to());

                    let new_distance_to_v = distance.clone() + edge.weight().clone();
                    forward_heap.push(Label::with_weight(edge.to(), new_distance_to_v));
                }
            }
        } else {
            forward_settle = true;

            // Extract a label from the forward Priority Queue
            let Label { id, distance } = backward_heap.pop().unwrap();

            //			debug!("Backward Extracted {} at distance {:?}", _id, _distance);
            // Settle the vertex
            settled_vertices_backward.insert(id, distance.clone());

            // Iterate over all the edges
            for edge in graph.inbound_edges(id) {
                //				debug!(
                //					"iterating over edge to {}, weight {:?}",
                //					edge.to(),
                //					edge.weight()
                //				);

                let mut update_best_path_distance = |_distance_to_v: &W| {
                    let new_distance_to_v =
                        distance.clone() + edge.weight().clone() + _distance_to_v.clone();

                    if new_distance_to_v < best_path_distance.clone() {
                        best_path_distance = new_distance_to_v;
                    }
                };

                if settled_vertices_forward.contains_key(&edge.to()) {
                    update_best_path_distance(settled_vertices_forward.get(&edge.to()).unwrap());
                } else if forward_heap.contains(&edge.to()) {
                    update_best_path_distance(
                        &forward_heap.best_known_distance_to_id(&edge.to()).unwrap(),
                    );
                }

                // If the vertex hasn't been settled yet
                if !settled_vertices_backward.contains_key(&edge.to()) {
                    //					debug!("Checking edge: {}", edge.to());

                    let new_distance_to_v = distance.clone() + edge.weight().clone();
                    backward_heap.push(Label::with_weight(edge.to(), new_distance_to_v));
                }
            }
        }
    }

    //	debug!(
    //		"Could not find a path, exiting, forward heap size = {}, back heap size = {}",
    //		forward_heap.len(),
    //		backward_heap.len()
    //	);
    None
}

// These tests are taken from https://github.com/easbar/fast_paths/blob/master/src/dijkstra.rs
// Thanks are due to easbar (Andreas Barth)!
//
// #[cfg(test)]
// mod tests {
// 	use super::*;
//
// 	#[test]
// 	fn simple_path() {
// 		//      7 -> 8 -> 9
// 		//      |         |
// 		// 0 -> 5 -> 6 -  |
// 		// |         |  \ |
// 		// 1 -> 2 -> 3 -> 4
// 		let mut g = Graph::<SingleWeightEdge<UnitWeight<usize>>>::new(10);
// 		g.add_edge(
// 			0,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(1, UnitWeight::<usize>::new(1)),
// 		);
// 		g.add_edge(
// 			1,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(2, UnitWeight::<usize>::new(1)),
// 		);
// 		g.add_edge(
// 			2,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(3, UnitWeight::<usize>::new(1)),
// 		);
// 		g.add_edge(
// 			3,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(4, UnitWeight::<usize>::new(20)),
// 		);
// 		g.add_edge(
// 			0,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(5, UnitWeight::<usize>::new(5)),
// 		);
// 		g.add_edge(
// 			5,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(6, UnitWeight::<usize>::new(1)),
// 		);
// 		g.add_edge(
// 			6,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(4, UnitWeight::<usize>::new(20)),
// 		);
// 		g.add_edge(
// 			6,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(3, UnitWeight::<usize>::new(20)),
// 		);
// 		g.add_edge(
// 			5,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(7, UnitWeight::<usize>::new(5)),
// 		);
// 		g.add_edge(
// 			7,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(8, UnitWeight::<usize>::new(1)),
// 		);
// 		g.add_edge(
// 			8,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(9, UnitWeight::<usize>::new(1)),
// 		);
// 		g.add_edge(
// 			9,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(4, UnitWeight::<usize>::new(1)),
// 		);
//
// 		assert_no_path(dijkstra::<SingleWeightEdge<UnitWeight<usize>>, UnitWeight<usize>>(
// 			&g, 4, 0,
// 		));
// 		assert_no_path(dijkstra::<SingleWeightEdge<UnitWeight<usize>>, UnitWeight<usize>>(
// 			&g, 4, 4,
// 		));
//
// 		assert_path(
// 			dijkstra::<SingleWeightEdge<UnitWeight<usize>>, UnitWeight<usize>>(&g, 6, 3),
// 			20,
// 		);
// 		assert_path(
// 			dijkstra::<SingleWeightEdge<UnitWeight<usize>>, UnitWeight<usize>>(&g, 1, 4),
// 			22,
// 		);
// 		assert_path(
// 			dijkstra::<SingleWeightEdge<UnitWeight<usize>>, UnitWeight<usize>>(&g, 0, 4),
// 			13,
// 		);
//
// 		assert_no_path(bidirectional_dijkstra::<
// 			SingleWeightEdge<UnitWeight<usize>>,
// 			UnitWeight<usize>,
// 		>(&g, 4, 0));
// 		assert_no_path(bidirectional_dijkstra::<
// 			SingleWeightEdge<UnitWeight<usize>>,
// 			UnitWeight<usize>,
// 		>(&g, 4, 4));
//
// 		assert_path(
// 			bidirectional_dijkstra::<SingleWeightEdge<UnitWeight<usize>>, UnitWeight<usize>>(&g, 6, 3),
// 			20,
// 		);
// 		assert_path(
// 			bidirectional_dijkstra::<SingleWeightEdge<UnitWeight<usize>>, UnitWeight<usize>>(&g, 1, 4),
// 			22,
// 		);
// 		assert_path(
// 			bidirectional_dijkstra::<SingleWeightEdge<UnitWeight<usize>>, UnitWeight<usize>>(&g, 0, 4),
// 			13,
// 		);
//
// 		// assert_path(&mut d, &g, 4, 4, 0, vec![4]);
// 		// assert_path(&mut d, &g, 6, 3, 20, vec![6, 3]);
// 		// assert_path(&mut d, &g, 1, 4, 22, vec![1, 2, 3, 4]);
// 		// assert_path(&mut d, &g, 0, 4, 13, vec![0, 5, 7, 8, 9, 4]);
// 	}
//
// 	#[test]
// 	fn go_around() {
// 		// 0 -> 1
// 		// |    |
// 		// 2 -> 3
// 		let mut g = Graph::<SingleWeightEdge<UnitWeight<usize>>>::new(4);
// 		g.add_edge(
// 			0,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(1, UnitWeight::<usize>::new(10)),
// 		);
// 		g.add_edge(
// 			0,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(2, UnitWeight::<usize>::new(1)),
// 		);
// 		g.add_edge(
// 			2,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(3, UnitWeight::<usize>::new(1)),
// 		);
// 		g.add_edge(
// 			3,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(1, UnitWeight::<usize>::new(1)),
// 		);
//
// 		assert_path(
// 			dijkstra::<SingleWeightEdge<UnitWeight<usize>>, UnitWeight<usize>>(&g, 0, 1),
// 			3,
// 		);
// 		assert_path(
// 			bidirectional_dijkstra::<SingleWeightEdge<UnitWeight<usize>>, UnitWeight<usize>>(&g, 0, 1),
// 			3,
// 		);
// 	}
//
// 	#[test]
// 	fn avoid_node() {
// 		// 0 -> 1 -> 2
// 		// |         |
// 		// 3 -> 4 -> 5
// 		let mut g = Graph::<SingleWeightEdge<UnitWeight<usize>>>::new(6);
// 		g.add_edge(
// 			0,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(1, UnitWeight::<usize>::new(1)),
// 		);
// 		g.add_edge(
// 			1,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(2, UnitWeight::<usize>::new(1)),
// 		);
// 		g.add_edge(
// 			0,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(3, UnitWeight::<usize>::new(10)),
// 		);
// 		g.add_edge(
// 			3,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(4, UnitWeight::<usize>::new(1)),
// 		);
// 		g.add_edge(
// 			4,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(5, UnitWeight::<usize>::new(1)),
// 		);
// 		g.add_edge(
// 			5,
// 			SingleWeightEdge::<UnitWeight<usize>>::new(2, UnitWeight::<usize>::new(1)),
// 		);
//
// 		assert_path(
// 			dijkstra::<SingleWeightEdge<UnitWeight<usize>>, UnitWeight<usize>>(&g, 0, 2),
// 			2,
// 		);
// 		assert_path(
// 			bidirectional_dijkstra::<SingleWeightEdge<UnitWeight<usize>>, UnitWeight<usize>>(&g, 0, 2),
// 			2,
// 		);
// 	}
//
// 	fn assert_no_path(observed_weight: Option<UnitWeight<usize>>) {
// 		assert_eq!(observed_weight, None);
// 	}
//
// 	fn assert_path(observed_weight: Option<UnitWeight<usize>>, correct_weight: usize) {
// 		if let Some(wt) = observed_weight {
// 			assert_eq!(wt, UnitWeight::<usize>::new(correct_weight));
// 		}
// 	}
// }
