use crate::data_structures::graphs::weight::WeightTrait;

// ---------------------------- Traits ----------------------------

// All edges need to have a destination vertex, and need to be mutable.
pub trait EdgeTrait: Clone {
    type Weight;

    fn new(dst: u32, weight: Self::Weight) -> Self;
    fn to(&self) -> u32;
    fn weight(&self) -> &Self::Weight;
    fn change_dst(&mut self, from: u32); // Needs to be mutable
    fn reduce_weight_from_edge_if_lower(&mut self, edge: &Self);
}

pub trait HopLengthTrait: EdgeTrait {
    fn new(dst: u32, weight: Self::Weight, hop_length: u32) -> Self;
    fn hop_length(&self) -> u32;
}

pub trait RankTrait: EdgeTrait {
    fn new(dst: u32, weight: Self::Weight, rank: Option<u32>) -> Self;
    fn new_unranked(dst: u32, weight: Self::Weight) -> Self;
    fn set_rank(&mut self, new_rank: Option<u32>);
    fn rank(&self) -> Option<u32>;
}

// -------------------------- Implementations ----------------------

// Edge containing a single weight struct and a separate field for hop length
// Used in the CH implementation
#[derive(Copy, Clone, Debug)]
pub struct WeightedEdge<W>
where
    W: WeightTrait,
{
    dst: u32,
    weight: W,
}

impl<W> EdgeTrait for WeightedEdge<W>
where
    W: WeightTrait,
{
    type Weight = W;

    fn new(dst: u32, weight: W) -> WeightedEdge<W> {
        WeightedEdge { dst, weight }
    }

    fn to(&self) -> u32 {
        self.dst
    }

    fn weight(&self) -> &W {
        &self.weight
    }

    fn change_dst(&mut self, dst: u32) {
        self.dst = dst;
    }

    fn reduce_weight_from_edge_if_lower(&mut self, edge: &WeightedEdge<W>) {
        if edge.weight() < &self.weight {
            self.weight = edge.weight().clone();
        }
    }
}

// // Edge containing a single weight struct and a separate field for hop length
// // Used in the CH implementation
#[derive(Copy, Clone, Debug)]
pub struct WeightedEdgeWithHopLength<W>
where
    W: WeightTrait,
{
    dst: u32,
    weight: W,
    hop_length: u32,
}

impl<W> EdgeTrait for WeightedEdgeWithHopLength<W>
where
    W: WeightTrait,
{
    type Weight = W;

    fn new(dst: u32, weight: W) -> WeightedEdgeWithHopLength<W> {
        WeightedEdgeWithHopLength {
            dst,
            weight,
            hop_length: 1,
        }
    }

    fn to(&self) -> u32 {
        self.dst
    }

    fn weight(&self) -> &W {
        &self.weight
    }

    fn change_dst(&mut self, dst: u32) {
        self.dst = dst;
    }

    fn reduce_weight_from_edge_if_lower(&mut self, edge: &WeightedEdgeWithHopLength<W>) {
        if edge.weight() < &self.weight {
            self.weight = edge.weight().clone();
        }
    }
}

impl<W> HopLengthTrait for WeightedEdgeWithHopLength<W>
where
    W: WeightTrait,
{
    fn new(dst: u32, weight: W, hop_length: u32) -> WeightedEdgeWithHopLength<W> {
        WeightedEdgeWithHopLength {
            dst,
            weight,
            hop_length,
        }
    }

    fn hop_length(&self) -> u32 {
        self.hop_length
    }
}

// Edge containing a single weight struct and a rank
// Used in the EH implementation
#[derive(Copy, Clone, Debug)]
pub struct RankedWeightedEdge<W>
where
    W: WeightTrait,
{
    dst: u32,
    weight: W,
    // ASSUMPTION: THE NUMBER OF EDGES OF A GRAPH < 2^32-1
    // Number of edges of a graph >> number of vertices
    rank: Option<u32>, // Demian's implementation uses u32::max() to mark edges 'unranked'.
}

impl<W> EdgeTrait for RankedWeightedEdge<W>
where
    W: WeightTrait,
{
    type Weight = W;

    fn new(dst: u32, weight: W) -> RankedWeightedEdge<W> {
        RankedWeightedEdge {
            dst,
            weight,
            rank: None,
        }
    }

    fn to(&self) -> u32 {
        self.dst
    }

    fn weight(&self) -> &W {
        &self.weight
    }

    fn change_dst(&mut self, dst: u32) {
        self.dst = dst;
    }

    fn reduce_weight_from_edge_if_lower(&mut self, edge: &RankedWeightedEdge<W>) {
        if edge.weight() < &self.weight {
            self.weight = edge.weight().clone();
        }
    }
}

impl<W> RankTrait for RankedWeightedEdge<W>
where
    W: WeightTrait,
{
    fn new(dst: u32, weight: W, rank: Option<u32>) -> RankedWeightedEdge<W> {
        RankedWeightedEdge { dst, weight, rank }
    }

    fn new_unranked(dst: u32, weight: W) -> RankedWeightedEdge<W> {
        RankedWeightedEdge {
            dst,
            weight,
            rank: None,
        }
    }

    fn set_rank(&mut self, new_rank: Option<u32>) {
        self.rank = new_rank;
    }

    fn rank(&self) -> Option<u32> {
        self.rank
    }
}

// Edges can have multiple weights, compound edge weights, ....
// All that needs to be done is implement the EdgeTrait
