use crate::data_structures::graphs::edge::*;
use std::fmt::Debug;

/********************************************************
NOTE ON REDUNDANCY:
This file contains 2 implementations of graphs:

Only allow edges with hop length:
Graph: A double sided graph that stores both an adjacency list and a reverse adjacency list
SingleSideGraph: A simple adjacency list graph

*********************************************************/

// Limitations/ Assumptions made while implementing the Graph class:
// 1. The graph has a fixed number of vertices
// 2. Assume that the vertex IDs start from 0.
// 3. Max number of vertices can be the largest value of the unsigned data type.
// 4. Do NOT add multiple edges between the same pair of vertices.
// 	  The Dijkstra's implementation can handle duplicate edges, but the remove_edge() WILL bork.

pub trait GraphTrait<E>
where
    E: EdgeTrait,
{
    // For constructing the graph
    fn new(vertices: u32) -> Self
    where
        Self: Sized;
    fn new_empty() -> Self
    where
        Self: Sized;
    fn set_num_vertices(&mut self, vertices: u32);

    // Modifiers
    fn add_edge(&mut self, from: u32, edge: E);

    // Accessors
    fn num_vertices(&self) -> u32;
    fn num_edges(&self) -> u32;

    fn out_degree(&self, vertex: u32) -> u32;
    fn outbound_neighbours(&self, vertex: u32) -> Vec<u32>;
    fn outbound_edge(&self, vertex: u32, i: u32) -> Option<E>;
    fn outbound_edges(&self, vertex: u32) -> &Vec<E>;
    fn get_outbound_edge_between_vertices(&self, vertex_1: u32, vertex_2: u32) -> Option<E>;
}

pub trait DoubleSidedGraphTrait<E>: GraphTrait<E>
where
    E: EdgeTrait,
{
    fn in_degree(&self, vertex: u32) -> u32;
    fn inbound_neighbours(&self, vertex: u32) -> Vec<u32>;
    fn inbound_edge(&self, vertex: u32, i: u32) -> Option<E>;
    fn inbound_edges(&self, vertex: u32) -> &Vec<E>;
}

// The default implementation of the Graph is a double-sided graph,
// i.e. it stores both the forward and backward adjacency lsits.
#[derive(Clone)]
pub struct Graph<E>
where
    E: EdgeTrait + Debug,
{
    adj_list: Vec<Vec<E>>,
    reverse_adj_list: Vec<Vec<E>>,
}

impl<E> GraphTrait<E> for Graph<E>
where
    E: EdgeTrait + Debug,
{
    fn new(vertices: u32) -> Graph<E> {
        Graph {
            adj_list: vec![Vec::new(); vertices as usize],
            reverse_adj_list: vec![Vec::new(); vertices as usize],
        }
    }

    fn new_empty() -> Graph<E> {
        Graph {
            adj_list: Vec::new(),
            reverse_adj_list: Vec::new(),
        }
    }

    // CAUTION:
    // Use this function only immediately after empty_new()
    // Changing the number of vertices after edges have been added to it is not allowed
    fn set_num_vertices(&mut self, vertices: u32) {
        if self.adj_list.len() != 0 {
            panic!("Changing graph size of a non-empty graph is not allowed!");
        }

        self.adj_list.reserve(vertices as usize);
        self.reverse_adj_list.reserve(vertices as usize);
    }

    fn add_edge(&mut self, from: u32, mut edge: E) {
        debug_assert!(from < self.num_vertices()); // This assertion has caught _so many_ bugs.

        unsafe { (*self.adj_list.get_unchecked_mut(from as usize)).push(edge.clone()) };

        let dst_vertex = edge.to();
        edge.change_dst(from as u32);
        unsafe {
            (*self.reverse_adj_list.get_unchecked_mut(dst_vertex as usize)).push(edge);
        }
    }

    fn num_vertices(&self) -> u32 {
        self.adj_list.len() as u32
    }

    // EXPENSIVE FUNCTION:
    // Maybe: Maintain a variable for edge count in the struct?
    // Since we're using filter() and chained iterators extensively, it's difficult to maintain a count
    // of edges being added or deleted from the graph
    // Try to minimize the calls to this function
    fn num_edges(&self) -> u32 {
        let mut edge_count = 0;
        for v in 0..self.num_vertices() {
            edge_count += self.adj_list[v as usize].len() as u32;
        }
        edge_count
    }
    fn out_degree(&self, vertex: u32) -> u32 {
        unsafe { self.adj_list.get_unchecked(vertex as usize).len() as u32 }
    }

    // The following two functions are expensive.
    // Note: Optimizing the two functions below will need an alternative
    // representation of the graph like an edge list.
    // However, since degrees of most vertices is likely to be small,
    // I am assuming that this works well enough for now.
    fn outbound_neighbours(&self, from: u32) -> Vec<u32> {
        unsafe {
            self.adj_list
                .get_unchecked(from as usize)
                .iter()
                .filter(|x| x.to() != from)
                .map(|x| x.to())
                .collect()
        }
    }

    // Get the i^th edge of the _from vertex in the graph
    fn outbound_edge(&self, vertex: u32, i: u32) -> Option<E> {
        self.adj_list[vertex as usize].get(i as usize).cloned()
    }

    // pub fn outbound_upward_edges(&self, _vertex: u32) -> Vec<E> {
    // 	unsafe{	self.adj_list.get_unchecked(_vertex as usize).iter().filter(|x| x.is_upward()).collect() }
    // }

    fn outbound_edges(&self, vertex: u32) -> &Vec<E> {
        unsafe { &self.adj_list.get_unchecked(vertex as usize) }
    }

    fn get_outbound_edge_between_vertices(&self, vertex_1: u32, vertex_2: u32) -> Option<E> {
        None
    }
}

impl<E> DoubleSidedGraphTrait<E> for Graph<E>
where
    E: EdgeTrait + Debug,
{
    fn in_degree(&self, vertex: u32) -> u32 {
        unsafe { self.reverse_adj_list.get_unchecked(vertex as usize).len() as u32 }
    }

    // The following two functions are expensive.
    // Note: Optimizing the two functions below will need an alternative
    // representation of the graph like an edge list.
    // However, since degrees of most vertices is likely to be small,
    // I am assuming that this works well enough for now.
    fn inbound_neighbours(&self, from: u32) -> Vec<u32> {
        unsafe {
            self.reverse_adj_list
                .get_unchecked(from as usize)
                .iter()
                .filter(|x| x.to() != from)
                .map(|x| x.to())
                .collect()
        }
    }

    fn inbound_edge(&self, vertex: u32, i: u32) -> Option<E> {
        self.reverse_adj_list[vertex as usize]
            .get(i as usize)
            .cloned()
    }

    fn inbound_edges(&self, vertex: u32) -> &Vec<E> {
        unsafe { &self.reverse_adj_list.get_unchecked(vertex as usize) }
    }
}

impl<E> Graph<E>
where
    E: EdgeTrait + Debug,
{
    pub fn add_edge_or_reduce_weight(&mut self, from: u32, mut edge: E) {
        debug_assert!(from < self.num_vertices());

        // Does edge exist?
        let mut edge_found = false;

        // For all edges in the adj_list where the destination is the same as input edge
        for out_edge in self.adj_list[from as usize]
            .iter_mut()
            .filter(|e| e.to() == edge.to())
        {
            out_edge.reduce_weight_from_edge_if_lower(&edge);
            let dest = out_edge.to();
            for in_edge in self.reverse_adj_list[dest as usize]
                .iter_mut()
                .filter(|e| e.to() == from)
            {
                in_edge.reduce_weight_from_edge_if_lower(&edge);
            }
            edge_found = true;
        }

        for in_edge in self.reverse_adj_list[edge.to() as usize]
            .iter_mut()
            .filter(|e| e.to() == from)
        {
            in_edge.reduce_weight_from_edge_if_lower(&edge);
            let dest = in_edge.to();
            for out_edge in self.adj_list[dest as usize]
                .iter_mut()
                .filter(|e| e.to() == edge.to())
            {
                out_edge.reduce_weight_from_edge_if_lower(&edge);
            }
            edge_found = true;
        }

        // If edge not found in the graph, add an edge
        if !edge_found {
            unsafe { (*self.adj_list.get_unchecked_mut(from as usize)).push(edge.clone()) };

            let dst_vertex = edge.to();
            edge.change_dst(from as u32);
            unsafe {
                (*self.reverse_adj_list.get_unchecked_mut(dst_vertex as usize)).push(edge);
            }
        }
    }

    // CAUTION! This function removes all edges from the given vertex to the to() vertex of the input edge.
    //	pub fn remove_all_edges_to_vertex(&mut self, _from: u32, _edge: E) {
    //		self.adj_list[_from as usize].retain(|x| x.to() != _edge.to());
    //		self.reverse_adj_list[_edge.to() as usize].retain(|x| x.to() != _from);
    //	}

    pub fn remove_all_incident_edges(&mut self, from: u32) {
        for e in self.adj_list[from as usize].iter() {
            self.reverse_adj_list[e.to() as usize].retain(|x| x.to() != from);
        }

        for e in self.reverse_adj_list[from as usize].iter() {
            self.adj_list[e.to() as usize].retain(|x| x.to() != from);
        }

        self.adj_list[from as usize].clear();
        self.reverse_adj_list[from as usize].clear();

        // assert_eq!(self.out_degree(from), 0);
        // assert_eq!(self.in_degree(from), 0);
    }
    pub fn print(&self) {
        println!("=================== FORWARD GRAPH ====================");
        for i in 0..self.adj_list.len() as u32 {
            print!("{} : ", i);
            for n in self.outbound_neighbours(i).iter() {
                print!("{} ", n);
            }
            println!("");
        }

        println!("=================== BACKWARD GRAPH ====================");
        for i in 0..self.reverse_adj_list.len() as u32 {
            print!("{} : ", i);
            for n in self.inbound_neighbours(i).iter() {
                print!("{} ", n);
            }
            println!("");
        }
    }

    pub fn print_edges(&self) {
        println!("=================== FORWARD GRAPH ====================");
        for i in 0..self.adj_list.len() as u32 {
            print!("{} : ", i);
            for e in self.outbound_edges(i).iter() {
                print!("{:?} ", e);
            }
            println!("");
        }

        println!("=================== BACKWARD GRAPH ====================");
        for i in 0..self.reverse_adj_list.len() as u32 {
            print!("{} : ", i);
            for e in self.inbound_edges(i).iter() {
                print!("{:?} ", e);
            }
            println!("");
        }
    }

    pub fn print_edge_counts(&self) {
        let mut forward_edge_count = 0;
        for i in self.adj_list.iter() {
            forward_edge_count += i.len();
        }

        let mut backward_edge_count = 0;
        for i in self.reverse_adj_list.iter() {
            backward_edge_count += i.len();
        }
        println!(
            "Forward edge count = {}, Backward edge count = {}",
            forward_edge_count, backward_edge_count
        );
    }
}

#[derive(Clone)]
pub struct SingleSideGraph<E>
where
    E: EdgeTrait + Debug,
{
    adj_list: Vec<Vec<E>>,
}

impl<E> GraphTrait<E> for SingleSideGraph<E>
where
    E: EdgeTrait + Debug,
{
    fn new(_vertices: u32) -> SingleSideGraph<E> {
        SingleSideGraph {
            adj_list: vec![Vec::new(); _vertices as usize],
        }
    }

    fn new_empty() -> SingleSideGraph<E> {
        SingleSideGraph {
            adj_list: Vec::new(),
        }
    }

    fn set_num_vertices(&mut self, _vertices: u32) {
        self.adj_list.reserve(_vertices as usize);
    }

    fn add_edge(&mut self, _from: u32, mut _edge: E) {
        debug_assert!(_from < self.num_vertices()); // This assertion has caught _so many_ bugs.

        unsafe { (*self.adj_list.get_unchecked_mut(_from as usize)).push(_edge.clone()) };
    }

    fn num_vertices(&self) -> u32 {
        self.adj_list.len() as u32
    }

    fn num_edges(&self) -> u32 {
        let mut edge_count: u32 = 0;
        for i in self.adj_list.iter() {
            edge_count += i.len() as u32;
        }
        edge_count
    }

    // TODO: Rename the following functions to get_edge() instead of get_outbound_edge(), since 'outbound' has no meaning in a single side graph

    fn out_degree(&self, _vertex: u32) -> u32 {
        unsafe { self.adj_list.get_unchecked(_vertex as usize).len() as u32 }
    }

    // The following two functions are expensive.
    // Note: Optimizing the two functions below will need an alternative
    // representation of the graph like an edge list.
    // However, since degrees of most vertices is likely to be small,
    // I am assuming that this works well enough for now.
    fn outbound_neighbours(&self, from: u32) -> Vec<u32> {
        unsafe {
            self.adj_list
                .get_unchecked(from as usize)
                .iter()
                .map(|x| x.to())
                .collect()
        }
    }

    // Get the i^th edge of the given vertex in the graph
    fn outbound_edge(&self, vertex: u32, i: u32) -> Option<E> {
        self.adj_list[vertex as usize].get(i as usize).cloned()
    }

    fn outbound_edges(&self, vertex: u32) -> &Vec<E> {
        unsafe { &self.adj_list.get_unchecked(vertex as usize) }
    }

    // Get the edge between two vertices in the graph
    // In case there are multiple edges, only the first match is returned
    // In case no edge is found, it returns None
    fn get_outbound_edge_between_vertices(&self, vertex_1: u32, vertex_2: u32) -> Option<E> {
        self.adj_list[vertex_1 as usize]
            .iter()
            .find(|&x| x.to() == vertex_2)
            .cloned()
    }
}

impl<E> SingleSideGraph<E>
where
    E: EdgeTrait + Debug,
{
    pub fn print(&self) {
        println!("=================== SingleSideGraph ====================");
        for i in 0..self.adj_list.len() as u32 {
            print!("{} : ", i);
            for n in self.outbound_neighbours(i).iter() {
                print!("{} ", n);
            }
            println!("");
        }
    }

    pub fn print_edges(&self) {
        println!("=================== SingleSideGraph ====================");
        for i in 0..self.adj_list.len() as u32 {
            print!("{} : ", i);
            for e in self.outbound_edges(i).iter() {
                print!("{:?} ", e);
            }
            println!("");
        }
    }
}

// Implements the Builder pattern for graphs, with methods for reading from dimacs and pbf files
//pub struct GraphBuilder<E> where E:EdgeTrait + Debug {
//	graph: Option<Graph<E>>
//}
//
//impl GraphBuilder<E> where E: EdgeTrait + Debug {
//	fn new() -> GraphBuilder<E> {
//		GraphBuilder {
//			graph: None
//		}
//	}
//
////	fn read_pbf_file(mut self, _filename: &String) -> GraphBuilder<E> {
////
////	}
//
//	fn read_dimacs_file(mut self, _filename: &String) -> GraphBuilder<E> {
//		unimplemented!()
//	}
//
//	fn build(&self) -> Result<Graph<E>, ()> {
//		unimplemented!()
//	}
//}
