use num::{Bounded, Zero};
use std::cmp::Ordering;
use std::ops::Add;

// The weightTrait must be implemented by all types that can be used as weights on the edges
// including histograms, functions, complex weights, everything.
pub trait WeightTrait: Clone + Ord + PartialEq {
    fn MAX() -> Self;
    fn MIN() -> Self;
    fn ZERO() -> Self;
}

// Create a UnitWeight with numeric trait implementations derived from the num crate.
#[derive(Clone, Debug)]
pub struct UnitWeight<W> {
    weight: W,
}

impl<W> UnitWeight<W>
where
    W: Clone + Ord + Eq + Add<Output = W> + Bounded + Zero,
{
    pub fn new(weight: W) -> UnitWeight<W> {
        UnitWeight { weight }
    }
}

impl<W> WeightTrait for UnitWeight<W>
where
    W: Clone + Ord + Eq + Add<Output = W> + Bounded + Zero,
{
    fn MAX() -> UnitWeight<W> {
        UnitWeight {
            weight: W::max_value(),
        }
    }

    fn MIN() -> UnitWeight<W> {
        UnitWeight {
            weight: W::min_value(),
        }
    }

    fn ZERO() -> UnitWeight<W> {
        UnitWeight { weight: W::zero() }
    }
}

impl<W> Add for UnitWeight<W>
where
    W: Clone + Ord + Eq + Add<Output = W> + Bounded + Zero,
{
    type Output = UnitWeight<W>;

    fn add(self, other: UnitWeight<W>) -> UnitWeight<W> {
        UnitWeight {
            weight: self.weight + other.weight,
        }
    }
}

impl<W> PartialOrd for UnitWeight<W>
where
    W: Clone + Ord + Eq + Add<Output = W> + Bounded + Zero,
{
    fn partial_cmp(&self, other: &UnitWeight<W>) -> Option<Ordering> {
        Some(self.weight.cmp(&other.weight))
    }
}

impl<W> Ord for UnitWeight<W>
where
    W: Clone + Ord + Eq + Add<Output = W> + Bounded + Zero,
{
    fn cmp(&self, other: &UnitWeight<W>) -> Ordering {
        self.weight.cmp(&other.weight)
    }
}

impl<W> PartialEq for UnitWeight<W>
where
    W: Clone + Ord + Eq + Add<Output = W> + Bounded + Zero,
{
    fn eq(&self, other: &UnitWeight<W>) -> bool {
        self.weight == other.weight
    }
}

impl<W> Eq for UnitWeight<W> where W: Clone + Ord + Eq + Add<Output = W> + Bounded + Zero {}

// Several types of edge weights can be defined here, compound types are allowed.
