// This file implements a minimal version of Rust standard library's binary heap

use core::mem::{swap, ManuallyDrop};
use core::ptr;

pub struct CustomBinaryHeap<T> {
	data: Vec<T>,
}

impl<T: Ord> CustomBinaryHeap<T> {
	fn new() -> CustomBinaryHeap<T> {
		CustomBinaryHeap { data: Vec::new() }
	}

	fn with_capacity(capacity: usize) -> CustomBinaryHeap<T> {
		CustomBinaryHeap {
			data: Vec::with_capacity(capacity),
		}
	}

	fn push(&mut self, item: T) {
		let old_len = self.len();
		self.data.push(item);
		self.sift_up(0, old_len);
	}

	fn is_empty(&self) -> bool {
		self.data.is_empty()
	}

	fn len(&self) -> usize {
		self.data.len()
	}

	fn pop(&mut self) -> Option<T> {
		self.data.pop().map(|mut item| {
			if !self.is_empty() {
				swap(&mut item, &mut self.data[0]);
				self.sift_down_to_bottom(0);
			}
			item
		})
	}

	fn change_priority(&mut self, item: T) {}

	fn sift_up(&mut self, start: usize, pos: usize) -> usize {
		unsafe {
			let mut hole = Hole::new(&mut self.data, pos);

			while hole.pos() > start {
				let parent = (hole.pos() - 1) / 2;
				if hole.element() <= hole.get(parent) {
					break;
				}
				hole.move_to(parent);
			}
			hole.pos()
		}
	}

	/// Take an element at `pos` and move it down the heap,
	/// while its children are larger.
	fn sift_down_range(&mut self, pos: usize, end: usize) {
		unsafe {
			let mut hole = Hole::new(&mut self.data, pos);
			let mut child = 2 * pos + 1;
			while child < end {
				let right = child + 1;
				// compare with the greater of the two children
				if right < end && !(hole.get(child) > hole.get(right)) {
					child = right;
				}
				// if we are already in order, stop.
				if hole.element() >= hole.get(child) {
					break;
				}
				hole.move_to(child);
				child = 2 * hole.pos() + 1;
			}
		}
	}

	fn sift_down(&mut self, pos: usize) {
		let len = self.len();
		self.sift_down_range(pos, len);
	}

	/// Take an element at `pos` and move it all the way down the heap,
	/// then sift it up to its position.
	///
	/// Note: This is faster when the element is known to be large / should
	/// be closer to the bottom.
	fn sift_down_to_bottom(&mut self, mut pos: usize) {
		let end = self.len();
		let start = pos;
		unsafe {
			let mut hole = Hole::new(&mut self.data, pos);
			let mut child = 2 * pos + 1;
			while child < end {
				let right = child + 1;
				// compare with the greater of the two children
				if right < end && !(hole.get(child) > hole.get(right)) {
					child = right;
				}
				hole.move_to(child);
				child = 2 * hole.pos() + 1;
			}
			pos = hole.pos;
		}
		self.sift_up(start, pos);
	}

	fn rebuild(&mut self) {
		let mut n = self.len() / 2;
		while n > 0 {
			n -= 1;
			self.sift_down(n);
		}
	}
}

/// Hole represents a hole in a slice i.e., an index without valid value
/// (because it was moved from or duplicated).
/// In drop, `Hole` will restore the slice by filling the hole
/// position with the value that was originally removed.
struct Hole<'a, T: 'a> {
	data: &'a mut [T],
	elt: ManuallyDrop<T>,
	pos: usize,
}

impl<'a, T> Hole<'a, T> {
	/// Create a new `Hole` at index `pos`.
	///
	/// Unsafe because pos must be within the data slice.
	#[inline]
	unsafe fn new(data: &'a mut [T], pos: usize) -> Self {
		debug_assert!(pos < data.len());
		// SAFE: pos should be inside the slice
		let elt = ptr::read(data.get_unchecked(pos));
		Hole {
			data,
			elt: ManuallyDrop::new(elt),
			pos,
		}
	}

	#[inline]
	fn pos(&self) -> usize {
		self.pos
	}

	/// Returns a reference to the element removed.
	#[inline]
	fn element(&self) -> &T {
		&self.elt
	}

	/// Returns a reference to the element at `index`.
	///
	/// Unsafe because index must be within the data slice and not equal to pos.
	#[inline]
	unsafe fn get(&self, index: usize) -> &T {
		debug_assert!(index != self.pos);
		debug_assert!(index < self.data.len());
		self.data.get_unchecked(index)
	}

	/// Move hole to new location
	///
	/// Unsafe because index must be within the data slice and not equal to pos.
	#[inline]
	unsafe fn move_to(&mut self, index: usize) {
		debug_assert!(index != self.pos);
		debug_assert!(index < self.data.len());
		let index_ptr: *const _ = self.data.get_unchecked(index);
		let hole_ptr = self.data.get_unchecked_mut(self.pos);
		ptr::copy_nonoverlapping(index_ptr, hole_ptr, 1);
		self.pos = index;
	}
}

impl<T> Drop for Hole<'_, T> {
	#[inline]
	fn drop(&mut self) {
		// fill the hole again
		unsafe {
			let pos = self.pos;
			ptr::copy_nonoverlapping(&*self.elt, self.data.get_unchecked_mut(pos), 1);
		}
	}
}
