// A priority queue where the elements are IDs from 0 to id_count -1 where id_count is a number that is set in the constructor
// The elements are sorted by integer keys

// Rust implementation of RoutingKit's id_queue.h queue
// Follows the same (confusing) variable names as used in routingkit
// They call it an (ID,Key) what is (Key,Value) in regular parlance

// The code is a cemetery of u32 <-> usize conversions, mainly because Rust insists on all indexing be usize
use std::fmt::Debug;
use std::ops::Add;
use std::u32;

use crate::data_structures::graphs::weight::WeightTrait;

// First element is ID, the second is Key
pub type IDKeyPair<W> = (u32, W);

pub struct MinIDQueue<W>
where
	W: WeightTrait + Add<Output = W> + Debug,
{
	id_pos: Vec<u32>,
	heap: Vec<IDKeyPair<W>>,
	heap_size: u32,
}

impl<W> MinIDQueue<W>
where
	W: WeightTrait + Add<Output = W> + Debug,
{
	const TREE_ARITY: u32 = 4;

	pub fn new(id_count: u32) -> MinIDQueue<W> {
		MinIDQueue {
			id_pos: vec![u32::max_value(); id_count as usize],
			heap: vec![(0, W::ZERO()); id_count as usize],
			heap_size: 0,
		}
	}

	pub fn is_empty(&self) -> bool {
		self.heap_size == 0
	}

	pub fn size(&self) -> u32 {
		self.heap_size
	}

	pub fn id_count(&self) -> u32 {
		self.id_pos.len() as u32
	}

	pub fn contains_id(&self, id: u32) -> bool {
		self.id_pos[id as usize] != u32::max_value()
	}

	pub fn clear(&mut self) {
		for i in 0..self.heap_size {
			self.id_pos[self.heap[i as usize].0 as usize] = u32::max_value()
		}
		self.heap_size = 0;
	}

	// why is this needed?
	pub fn swap(l: &mut MinIDQueue<W>, r: &mut MinIDQueue<W>) {
		std::mem::swap(&mut l.id_pos, &mut r.id_pos);
		std::mem::swap(&mut l.heap, &mut r.heap);
		std::mem::swap(&mut l.heap_size, &mut r.heap_size);
	}

	pub fn get_key(&self, id: u32) -> W {
		assert_ne!(self.id_pos[id as usize], u32::max_value());
		self.heap[self.id_pos[id as usize] as usize].1.clone()
	}

	pub fn peek(&self) -> Option<&IDKeyPair<W>> {
		self.heap.first()
	}

	pub fn pop(&mut self) -> IDKeyPair<W> {
		assert_ne!(self.is_empty(), true);
		self.heap_size -= 1;
		self.heap.swap(0, self.heap_size as usize);
		self.id_pos[self.heap[0].0 as usize] = 0;
		self.id_pos[self.heap[self.heap_size as usize].0 as usize] = u32::max_value();

		self.move_down_in_tree(0);
		self.heap[self.heap_size as usize].clone()
	}

	pub fn push(&mut self, p: IDKeyPair<W>) {
		let pos = self.heap_size;
		self.heap_size += 1;
		self.id_pos[p.0 as usize] = pos;
		self.heap[pos as usize] = p;
		self.move_up_in_tree(pos);
	}

	pub fn decrease_key(&mut self, p: IDKeyPair<W>) -> bool {
		let pos = self.id_pos[p.0 as usize];
		if self.heap[pos as usize].1 > p.1 {
			self.heap[pos as usize].1 = p.1;
			self.move_up_in_tree(pos);
			true
		} else {
			false
		}
	}

	pub fn increase_key(&mut self, p: IDKeyPair<W>) -> bool {
		let pos = self.id_pos[p.0 as usize];
		if self.heap[pos as usize].1 < p.1 {
			self.heap[pos as usize].1 = p.1;
			self.move_down_in_tree(pos);
			true
		} else {
			false
		}
	}

	fn move_up_in_tree(&mut self, pos: u32) {
		let mut _pos = pos;
		while _pos != 0 {
			let _parent = (_pos - 1) / MinIDQueue::<W>::TREE_ARITY;
			if self.heap[_parent as usize].1 > self.heap[_pos as usize].1 {
				self.heap.swap(_pos as usize, _parent as usize);
				self.id_pos.swap(
					self.heap[_pos as usize].0 as usize,
					self.heap[_parent as usize].0 as usize,
				);
			}
			_pos = _parent;
		}
	}

	fn move_down_in_tree(&mut self, pos: u32) {
		let arity = MinIDQueue::<W>::TREE_ARITY;
		let mut _pos = pos;
		loop {
			let first_child = arity * _pos + 1;
			if first_child >= self.heap_size {
				return;
			}

			let mut smallest_child = first_child;
			let mut c = first_child + 1;
			while c < std::cmp::min(arity * _pos + arity + 1, self.heap_size) {
				if self.heap[smallest_child as usize].1 > self.heap[c as usize].1 {
					smallest_child = c;
				}
				c += 1;
			}

			if self.heap[smallest_child as usize].1 >= self.heap[_pos as usize].1 {
				return;
			}

			self.heap.swap(_pos as usize, smallest_child as usize);
			self.id_pos.swap(
				self.heap[_pos as usize].0 as usize,
				self.heap[smallest_child as usize].0 as usize,
			);
			_pos = smallest_child;
		}
	}

	pub fn print(&self) {
		println!("{:?}", self.heap);
	}
}
