use std::collections::BinaryHeap;
use std::fmt::Debug;

use crate::data_structures::graphs::weight::WeightTrait;
use rustc_hash::FxHashMap;
use std::cmp::Ordering;

// A label is set for each vertex in the graph when running Dijkstra's
#[derive(Eq, PartialEq, Debug)]
pub struct Label<W>
where
	W: WeightTrait,
{
	pub id: u32,
	pub distance: W,
}

impl<W> Label<W>
where
	W: WeightTrait,
{
	pub fn with_weight(id: u32, distance: W) -> Label<W> {
		Label { id, distance }
	}
}

impl<W> Ord for Label<W>
where
	W: WeightTrait + Eq + PartialEq,
{
	fn cmp(&self, other: &Label<W>) -> Ordering {
		other.distance.cmp(&self.distance).then_with(|| self.id.cmp(&other.id))
	}
}

impl<W> PartialOrd for Label<W>
where
	W: WeightTrait + Eq + PartialEq,
{
	fn partial_cmp(&self, other: &Label<W>) -> Option<Ordering> {
		Some(self.cmp(other))
	}
}

// This file provides a queue that is a thin wrapper over the standard binary heap, but allows
// us to query if the binary heap contains a value. The implementation should make the following guarantees:
//
// 1. At a time, only a single node of a given ID is present in the heap.
// 2. For a node present in the heap, it is guaranteed that the distance associated with it is the best distance
//    that has been attempted to be inserted in the heap.
// 3. The heap allows quick check to see if a node with a given ID is present inside it.

pub struct LabelHeap<W>
where
	W: WeightTrait,
{
	heap: BinaryHeap<Label<W>>,
	membership_set: FxHashMap<u32, W>,
}

impl<W> LabelHeap<W>
where
	W: WeightTrait,
{
	pub fn new() -> LabelHeap<W> {
		LabelHeap {
			heap: BinaryHeap::<Label<W>>::new(),
			membership_set: FxHashMap::<u32, W>::default(),
		}
	}

	pub fn is_empty(&self) -> bool {
		return self.membership_set.is_empty();
	}

	pub fn contains(&self, _id: &u32) -> bool {
		return self.membership_set.contains_key(_id);
	}

	pub fn clear(&mut self) {
		self.heap.clear();
		self.membership_set.clear();
	}

	pub fn best_known_distance_to_id(&self, _id: &u32) -> Option<W> {
		self.membership_set.get(&_id).cloned()
	}

	pub fn len(&self) -> u32 {
		self.heap.len() as u32
	}

	pub fn peek(&self) -> Option<&Label<W>> {
		self.heap.peek()
	}

	pub fn pop(&mut self) -> Option<Label<W>> {
		match self.heap.pop() {
			Some(_val) => {
				self.membership_set.remove(&_val.id);
				return Some(_val);
			}

			None => return None,
		}
	}

	// The heap maintains only the best known distance to a given vertex ID
	pub fn push(&mut self, label: Label<W>) {
		// println!("In the push function");

		// If the vertex is already in the heap
		match self.membership_set.get(&label.id) {
			Some(existing_distance) => {
				if existing_distance > &label.distance {
					// println!("GOt the distance,  inserting to the data structures");
					self.membership_set.remove(&label.id);
					self.membership_set.insert(label.id, label.distance.clone());
					self.heap.push(label);
				} else {
					// println!("GOt the distance, but existing distance < new distance");
				}
			}
			None => {
				self.membership_set.insert(label.id, label.distance.clone());
				self.heap.push(label);
			}
		}
	}
}
