use std::cmp::max;
use std::error::Error;

use crate::data_structures::graphs::edge::{
    EdgeTrait, RankTrait, RankedWeightedEdge, WeightedEdge, WeightedEdgeWithHopLength,
};
use crate::data_structures::graphs::graph::{Graph, GraphTrait};
use crate::data_structures::graphs::weight::UnitWeight;

use csv::ReaderBuilder;
use std::collections::BTreeMap;

type Record = (char, u32, u32, u32);

// NOTE: Self-loops in the graph are removed before creating the graph.
pub fn build_given_size(
    filename: String,
    num_vertices: u32,
) -> Result<Graph<WeightedEdge<UnitWeight<u32>>>, Box<dyn Error>> {
    println!("In the build function.");
    let mut reader = ReaderBuilder::new().delimiter(b' ').from_path(filename)?;
    let mut graph = Graph::<WeightedEdge<UnitWeight<u32>>>::new(num_vertices);

    println!("Starting to read records.. ");
    for row in reader.deserialize() {
        let record: Record = row?;
        if record.0 == 'a' && record.1 != record.2 {
            // It's okay to unwrap() since the DIMACS format guarantees it.
            graph.add_edge(
                record.1,
                WeightedEdge::<UnitWeight<u32>>::new(record.2, UnitWeight::<u32>::new(record.3)),
            );
        }
    }

    println!("Done reading records...");

    Ok(graph)
}

pub fn build_from_file(
    filename: String,
) -> Result<Graph<WeightedEdge<UnitWeight<u32>>>, Box<dyn Error>> {
    println!("In the build function.");
    let mut reader = ReaderBuilder::new().delimiter(b' ').from_path(filename)?;
    let mut vec_record = BTreeMap::<(u32, u32), u32>::new(); // This is used to avoid duplicate edges
    let mut num_vertices = 0;

    println!("Starting to read records.. ");
    for row in reader.deserialize() {
        let record: Record = row?;
        if record.0 == 'a' {
            if record.1 != record.2 {
                // Remove self edges
                vec_record.insert((record.1, record.2), record.3);

                // Count the number of vertices by finding the maximum vertex ID referred in the records
                if num_vertices < record.1 || num_vertices < record.2 {
                    num_vertices = max(record.1, record.2);
                }
            }
        }
    }
    println!("Done reading records...");

    println!(
        "Creating the graph of {} vertices and {} edges now...",
        num_vertices,
        vec_record.len()
    );
    let mut graph = Graph::<WeightedEdge<UnitWeight<u32>>>::new(num_vertices + 1);
    for ((source, dest), weight) in vec_record {
        graph.add_edge(
            source,
            WeightedEdge::<UnitWeight<u32>>::new(dest, UnitWeight::<u32>::new(weight)),
        );
        // println!("Adding edge {} -> {}, {}", r.1, r.2, r.3);
    }

    println!("Graph created.");
    Ok(graph)
}

// NOTE: Self-loops in the graph are removed before creating the graph.
pub fn build_given_size_with_hop_length(
    filename: String,
    num_vertices: u32,
) -> Result<Graph<WeightedEdgeWithHopLength<UnitWeight<u32>>>, Box<dyn Error>> {
    println!("In the build function.");
    let mut reader = ReaderBuilder::new().delimiter(b' ').from_path(filename)?;
    let mut graph = Graph::<WeightedEdgeWithHopLength<UnitWeight<u32>>>::new(num_vertices);

    println!("Starting to read records.. ");
    for row in reader.deserialize() {
        let record: Record = row?;
        if record.0 == 'a' && record.1 != record.2 {
            // It's okay to unwrap() since the DIMACS format guarantees it.
            graph.add_edge(
                record.1,
                WeightedEdgeWithHopLength::<UnitWeight<u32>>::new(
                    record.2,
                    UnitWeight::<u32>::new(record.3),
                ),
            );
        }
    }

    println!("Done reading records...");

    Ok(graph)
}

pub fn build_from_file_with_hop_length(
    filename: String,
) -> Result<Graph<WeightedEdgeWithHopLength<UnitWeight<u32>>>, Box<dyn Error>> {
    println!("In the build function.");
    let mut reader = ReaderBuilder::new().delimiter(b' ').from_path(filename)?;
    let mut vec_record = BTreeMap::<(u32, u32), u32>::new(); // This is used to avoid duplicate edges
    let mut num_vertices = 0;

    println!("Starting to read records.. ");
    for row in reader.deserialize() {
        let record: Record = row?;
        if record.0 == 'a' {
            if record.1 != record.2 {
                // Remove self edges
                vec_record.insert((record.1, record.2), record.3);

                // Count the number of vertices by finding the maximum vertex ID referred in the records
                if num_vertices < record.1 || num_vertices < record.2 {
                    num_vertices = max(record.1, record.2);
                }
            }
        }
    }
    println!("Done reading records...");

    println!(
        "Creating the graph of {} vertices and {} edges now...",
        num_vertices,
        vec_record.len()
    );
    let mut graph = Graph::<WeightedEdgeWithHopLength<UnitWeight<u32>>>::new(num_vertices + 1);
    for ((source, dest), weight) in vec_record {
        graph.add_edge(
            source,
            WeightedEdgeWithHopLength::<UnitWeight<u32>>::new(dest, UnitWeight::<u32>::new(weight)),
        );
        // println!("Adding edge {} -> {}, {}", r.1, r.2, r.3);
    }

    println!("Graph created.");
    Ok(graph)
}
